#!/user/bin/env python
# -*- coding: utf-8 -*-
"""
Copyright (C) 2024 Galaxy Technologies
Licensed under the Apache License, Version 2.0
"""
import math
import numpy as np
from pricelib.common.utilities.enums import CallPut
from pricelib.common.pricing_engine_base import QuadEngine
from pricelib.common.time import global_evaluation_date


class QuadAutoCallEngine(QuadEngine):
    """Autocall Note(二元小雪球) 数值积分定价引擎"""

    # pylint: disable=too-many-locals
    def calc_present_value(self, prod, t=None, spot=None):
        """计算现值
        Args:
            prod: Product产品对象
            t: datetime.date，估值日; 如果是None，则使用全局估值日globalEvaluationDate
            spot: float，估值日标的价格，如果是None，则使用随机过程的当前价格
        Returns: float，现值
        """
        calculate_date = global_evaluation_date() if t is None else t
        _maturity = (prod.end_date - calculate_date).days / prod.annual_days.value
        _maturity_business_days = prod.trade_calendar.business_days_between(calculate_date,
                                                                            prod.end_date) / prod.t_step_per_year
        obs_dates = prod.obs_dates.count_business_days(calculate_date)
        obs_dates = np.array([num / prod.t_step_per_year for num in obs_dates if num >= 0])
        calculate_start_diff = (calculate_date - prod.start_date).days
        pay_dates = prod.pay_dates.count_calendar_days(prod.start_date)
        pay_dates = np.array([num / prod.annual_days.value for num in pay_dates if num >= calculate_start_diff])
        assert len(obs_dates) == len(pay_dates), f"Error: {prod}的观察日和付息日长度不一致"
        diff_obs_pay_dates = np.array([(prod.pay_dates.date_schedule[i] - prod.obs_dates.date_schedule[i]).days
                                       for i in range(len(obs_dates))]) / prod.annual_days.value
        # 经过估值日截断列表，例如prod.barrier_out有22个，存续一年时估值，_barrier_out只有12个
        _barrier_out = prod.barrier_out[-len(obs_dates):].copy()
        _coupon_out = prod.coupon_out[-len(obs_dates):].copy()

        if spot is None:
            spot = self.process.spot()

        r = self.process.interest(_maturity)
        q = self.process.div(_maturity)
        vol = self.process.vol(_maturity, spot)

        self._check_method_params()
        self.set_quad_params(r=r, q=q, vol=vol)
        # 初始化fft的对数价格向量及边界的对数价格向量
        self.init_grid(spot, vol, _maturity_business_days)
        s_vec = np.exp(self.ln_s_vec)

        if obs_dates[0] == 0:
            # 如果估值日就是敲出观察日
            if (prod.callput == CallPut.Call and spot >= _barrier_out[0]) or (
                    prod.callput == CallPut.Put and spot <= _barrier_out[0]):  # 发生敲出
                return prod.s0 * (prod.margin_lvl + _coupon_out[0] * pay_dates[
                        0]) * math.exp(-r * diff_obs_pay_dates[0])
            else:  # 没有发生敲出，则积分迭代步数少一次
                s0_dt = obs_dates[1]
                dt_vec = np.diff(obs_dates)[1:]
        else:
            s0_dt = obs_dates[0]
            dt_vec = np.diff(obs_dates)

        self.backward_steps = dt_vec.size
        v_grid = np.zeros(shape=(self.n_points, self.backward_steps + 2))
        for step in range(self.backward_steps + 1, 0, -1):
            y = self.ln_s_vec
            current_barrier = np.searchsorted(s_vec, _barrier_out[step - 1], side='right')  # 行权价在价格格点上的对应格点

            if step == self.backward_steps + 1:
                if prod.callput == CallPut.Call:
                    v_grid[current_barrier:, -1] = (prod.margin_lvl + _coupon_out[-1] * pay_dates[-1]) * prod.s0
                    v_grid[:current_barrier, -1] = (prod.margin_lvl + prod.coupon_div * pay_dates[-1]) * prod.s0
                elif prod.callput == CallPut.Put:
                    v_grid[:current_barrier, -1] = (prod.margin_lvl + _coupon_out[-1] * pay_dates[-1]) * prod.s0
                    v_grid[current_barrier:, -1] = (prod.margin_lvl + prod.coupon_div * pay_dates[-1]) * prod.s0
            else:
                if prod.callput == CallPut.Call:
                    x = self.ln_s_vec[:current_barrier]
                    v_grid[:current_barrier, step] = self.fft_step_backward(x, y, v_grid[:, step + 1], dt_vec[step - 1])
                    v_grid[current_barrier:, step] = prod.s0 * (prod.margin_lvl + _coupon_out[step - 1] * pay_dates[
                        step - 1]) * math.exp(-r * diff_obs_pay_dates[step - 1])

                elif prod.callput == CallPut.Put:
                    x = self.ln_s_vec[current_barrier:]
                    v_grid[current_barrier:, step] = self.fft_step_backward(x, y, v_grid[:, step + 1], dt_vec[step - 1])
                    v_grid[:current_barrier, step] = prod.s0 * (prod.margin_lvl + _coupon_out[step - 1] * pay_dates[
                        step - 1]) * math.exp(-r * diff_obs_pay_dates[step - 1])

        x = np.array([np.log(spot)])
        value = self.fft_step_backward(x, self.ln_s_vec, v_grid[:, 1], s0_dt)[0]
        return value



#!/user/bin/env python
# -*- coding: utf-8 -*-
"""
Copyright (C) 2024 Galaxy Technologies
Licensed under the Apache License, Version 2.0
"""
import numpy as np
from pricelib.common.utilities.enums import CallPut, ProcessType, ExerciseType
from pricelib.common.processes import StochProcessBase
from pricelib.common.pricing_engine_base import AnalyticEngine
from pricelib.common.time import global_evaluation_date


class AnalyticHestonVanillaEngine(AnalyticEngine):
    """Heston半闭式解定价引擎"""

    def __init__(self, stoch_process: StochProcessBase):
        """
        初始化Heston半闭式解定价引擎
        Args:
            stoch_process: HestonProcess1D随机过程
        """
        assert stoch_process.process_type == ProcessType.Heston, "Error: Heston半闭式解定价引擎只能使用HestonProcess1D随机过程"
        super().__init__(stoch_process)
        # 以下属性指向需要定价的产品，由calc_present_value方法设置
        self.prod = None

    def calc_present_value(self, prod, t=None, spot=None):
        """计算现值
        Args:
            prod: Product产品对象
            t: datetime.date，估值日; 如果是None，则使用全局估值日globalEvaluationDate
            spot: float，估值日标的价格，如果是None，则使用随机过程的当前价格
        Returns: float，现值
        """
        assert prod.exercise_type == ExerciseType.European, "Error: 欧式期权的行权类型只能是European"
        self.prod = prod
        calculate_date = global_evaluation_date() if t is None else t
        tau = (prod.end_date - calculate_date).days / prod.annual_days.value
        if spot is None:
            spot = self.process.spot()
        r = self.process.interest(tau)
        q = self.process.div(tau)
        p_value1 = self.cal_p_value(prod.strike, tau, spot, r, q, self.process.v0, self.process.var_theta,
                                    self.process.var_rho, self.process.var_kappa, self.process.var_vol, 1)
        p_value2 = self.cal_p_value(prod.strike, tau, spot, r, q, self.process.v0, self.process.var_theta,
                                    self.process.var_rho, self.process.var_kappa, self.process.var_vol, 2)
        if prod.callput == CallPut.Call:
            price = spot * np.exp(-q * tau) * p_value1 - prod.strike * np.exp(-r * tau) * p_value2
        elif prod.callput == CallPut.Put:
            price = prod.strike * np.exp(-r * tau) * (1 - p_value2) - spot * np.exp(-q * tau) * (1 - p_value1)
        return price

    # pylint: disable=invalid-name, too-many-arguments, too-many-locals
    def cal_p_value(self, K, t, s0, r, q, v0, theta, rho, kappa, sigma, j_type, deg=64, m=250):
        """
        Heston猜解Pj，j=1或2，对应BSM解析解中的N(d1)或N(d2)，通过数值积分计算
        """

        def integral_func(phi):
            """被积函数(复数)"""
            return self.char_fun(t, s0, r, q, v0, theta, rho, kappa, sigma, j_type, phi) * np.exp(
                -1j * phi * np.log(K)) / (1j * phi)
            # 初始化高斯-勒让德积分，deg是积分划分节点，愈多精度愈高。x是高斯点，w是高斯点对应的权重

        x, w = np.polynomial.legendre.leggauss(deg)
        x = (x + 1) * 0.5 * m
        return 0.5 + (1 / np.pi) * np.sum(w * integral_func(x)).real * 0.5 * m  # 计算积分

    # pylint: disable=invalid-name, too-many-arguments, too-many-locals
    @staticmethod
    def char_fun(t, s0, r, q, v0, theta, rho, kappa, sigma, j_type, phi):
        """Heston特征函数"""
        if j_type == 1:
            u = 0.5
            b = kappa - rho * sigma
        elif j_type == 2:
            u = -0.5
            b = kappa
        else:
            raise ValueError(f'heston模型半闭式解中的j_type只能是1或2，当前为j_type={j_type}')
        a = kappa * theta
        x = np.log(s0)
        d = np.sqrt((rho * sigma * phi * 1j - b) ** 2 - sigma ** 2 * (2 * u * phi * 1j - phi ** 2))
        g = (b - rho * sigma * phi * 1j + d) / (b - rho * sigma * phi * 1j - d)
        capitalC = (r - q) * phi * 1j * t + (a / sigma ** 2) * (
                (b - rho * sigma * phi * 1j + d) * t - 2 * np.log((1 - g * np.exp(d * t)) / (1 - g)))
        capitalD = ((b - rho * sigma * phi * 1j + d) / sigma ** 2) * (1 - np.exp(d * t)) / (1 - g * np.exp(d * t))
        return np.exp(capitalC + capitalD * v0 + 1j * phi * x)



#!/user/bin/env python
# -*- coding: utf-8 -*-
"""
Copyright (C) 2024 Galaxy Technologies
Licensed under the Apache License, Version 2.0
"""
import numpy as np
from .utility import logging

try:
    import plotly.graph_objects as go
    import matplotlib.pyplot as plt
    import matplotlib.lines as mlines
except ModuleNotFoundError as e:
    logging.error(e)
    logging.warning('请安装plotly和matplotlib库,\nmatplotlib>=3.5.3, <=3.7.5;\nplotly>=5.16.1, <=5.22.0')
    exit(0)


def draw_greeks_surface(greeks_name, greeks_matrix, spot_range=None, show_plot=True):
    """绘制greeks曲面, spot_range是价格索引的区间
    Args:
        greeks_name: str, greeks名称, 如delta, gamma, theta等
        greeks_matrix: np.ndarray, 希腊值二维数组，行代表价格，列代表时间
        spot_range: tuple(int)=(spot_min, spot_max), 价格索引的区间, 默认为None, 即绘制全部价格范围的图像
        show_plot: bool, 是否显示图像, 默认为True
    Returns:
        fig: plotly.graph_objects.Figure对象，希腊值三维图
    """
    price_grid = np.array(range(1, 1 + greeks_matrix.shape[0]))
    if spot_range is not None:
        spot_min, spot_max = spot_range
        price_grid = price_grid[spot_min:spot_max + 1]
        greeks_matrix = greeks_matrix[spot_min:spot_max + 1, ]
    time_grid = np.array(range(greeks_matrix.shape[1]))
    fig = go.Figure(
        data=[go.Surface(z=greeks_matrix, x=time_grid, y=price_grid, colorscale='Turbo')], )  # Jet,Turbo, Rainbow
    # 轮廓线(可选)
    fig.update_traces(contours_z={  # 轮廓设置
        "show": True,  # 开启是否显示
        "usecolormap": True,  # 颜色设置
        "highlightcolor": "red",  # 高亮 mistyrose
        "project_z": True})
    # 设置标题、图片大小
    fig.update_layout(
        title=f'{greeks_name}-S-t surface',
        width=800, height=900,
        autosize=True,
        scene={
            'xaxis': {'title': '时间格点'},
            'yaxis': {'title': '标的价格格点'},
            'zaxis': {'title': greeks_name},
        },
        # template="plotly_dark",  # 深色主题
        margin={'l': 65, 'r': 50, 'b': 65, 't': 90},
        # margin=dict(r=0, l=0, t=0, b=0, pad=0)
    )
    if show_plot:
        # 将图保存到本地
        # pyo.plot(fig, filename=f'./greeks_surface_plot.html')
        fig.show()  # 显示图片(plotly在线)
    return fig


def draw_greeks_curve(delta_matrix, gamma_matrix=None, t=0, spot_range=None, show_plot=True):
    """绘制t时刻的greeks曲线
    Args:
        delta_matrix: np.ndarray, delta二维数组，行代表价格，列代表时间
        gamma_matrix: np.ndarray, gamma二维数组，行代表价格，列代表时间，默认为None，此时只绘制delta曲线
        t: int, 相对于起始日的时间索引
        spot_range: tuple(int)=(spot_min, spot_max), 价格索引的区间, 默认为None, 即绘制全部价格范围的图像
        show_plot: bool, 是否显示图像, 默认为True
    Returns:
        fig: plotly.graph_objects.Figure对象，希腊值曲线图
    """
    fig = go.Figure()
    price_grid = np.array(range(1, 1 + delta_matrix.shape[0]))
    if spot_range is not None:
        spot_min, spot_max = spot_range
        price_grid = price_grid[spot_min:spot_max + 1]
        delta_matrix = delta_matrix[spot_min:spot_max + 1, ]
        if gamma_matrix is not None:
            gamma_matrix = gamma_matrix[spot_min:spot_max + 1, ]

    delta_line = go.Scatter(
        x=price_grid,
        y=delta_matrix[:, t],
        mode='lines',
        name="Delta"
    )
    fig.add_trace(delta_line)
    if gamma_matrix is not None:
        gamma_line = go.Scatter(
            x=price_grid,
            y=gamma_matrix[:, t],
            mode='lines',
            name="Gamma"
        )
        fig.add_trace(gamma_line)
    fig.update_layout(title='Greeks', xaxis_title='Underlying Price',
                      yaxis_title='Greeks Value')  # 深色主题 , template="plotly_dark",

    if show_plot:
        # pyo.plot(fig, filename=f'./greeks_curve_plot.html')
        fig.show()
    return fig


def draw_greeks_contour(greeks_matrix, barrier_in, high_value=1.2, low_value=0.01, title=None):
    """绘制greeks等高线
    Args:
        greeks_matrix: np.ndarray, 希腊值二维数组，行代表价格，列代表时间
        barrier_in: int, barrier_in是敲入价索引
        high_value: float, greeks的上限
        low_value: float, greeks的下限
        title: str, 图表标题
    Returns: None
    """
    # 绘制Delta等高线图
    high_contour = plt.contour(greeks_matrix, levels=[high_value, ])
    # 清除之前的图像
    plt.clf()
    # 获取等高线的路径点
    for collection in high_contour.collections:
        paths = collection.get_paths()
        for path in paths:
            vertices = path.vertices
            # 筛选出上半边的点（未敲入）
            upper_half_vertices = vertices[vertices[:, 1] > barrier_in]
            # 绘制上半边的点（未敲入）
            plt.plot(upper_half_vertices[:, 0], upper_half_vertices[:, 1], 'orange')
    plt.contour(greeks_matrix, levels=[low_value, ], colors='royalblue')
    plt.xlabel('Days')
    plt.ylabel('Price')
    if title is None:
        title = f'雪球理论Delta区间: {low_value}-{high_value}'
    plt.title(title)
    # 创建虚拟线条对象用于图例
    high_line = mlines.Line2D([], [], color='orange', label=f'Delta={high_value}')
    low_line = mlines.Line2D([], [], color='royalblue', label=f'Delta={low_value}')
    plt.legend(handles=[high_line, low_line])
    # 控制y轴范围
    plt.ylim(barrier_in, 120)
    plt.rcParams['font.sans-serif'] = ['FangSong']  # 设置中文字体
    plt.show()



#!/user/bin/env python
# -*- coding: utf-8 -*-
"""
Copyright (C) 2024 Galaxy Technologies
Licensed under the Apache License, Version 2.0
"""
import datetime
from pricelib.common.utilities.utility import time_this, logging
from pricelib.common.utilities.enums import EngineType, ExerciseType, CallPut
from pricelib.common.product_base.option_base import OptionBase
from pricelib.common.time import CN_CALENDAR, AnnualDays, global_evaluation_date
from ...pricing_engines.analytic_engines.analytic_vanilla_european_engine import AnalyticVanillaEuEngine
from .vanilla_option import VanillaOption
from .underlying import UnderlyingAsset


class VanillaPortfolio(OptionBase):
    """香草期权组合
    可以组合出价差、跨式、宽跨式、风险反转、领口、鹰式、蝶式等香草期权组合"""

    def __init__(self, maturity=None, start_date=None, end_date=None, trade_calendar=CN_CALENDAR,
                 annual_days=AnnualDays.N365, t_step_per_year=243, s=None, r=None, q=None, vol=None):
        """构造函数
        Args:
            maturity: float，年化期限
            start_date: datetime.date，起始日
            end_date: datetime.date，到期日
            trade_calendar: 交易日历，Calendar类，默认为中国内地交易日历
            annual_days: int，每年的自然日数量
            t_step_per_year: int，每年的交易日数量
        初始化属性：
            underlying_list: 标的资产列表，[(underlying, position), ...]，其中underlying为UnderlyingAsset对象，position为持仓数量
            vanilla_list: 香草期权列表，[(vanilla, position), ...] ，其中vanilla为VanillaOption对象，position为持仓数量
        若提供了标的价格、无风险利率、分红/融券率、波动率，
            则默认使用解析解定价引擎 (BSM公式)
            s: float，标的价格
            r: float，无风险利率
            q: float，分红/融券率
            vol: float，波动率
        """
        super().__init__()
        logging.info("创建香草期权组合")
        self.trade_calendar = trade_calendar
        self.annual_days = annual_days
        self.t_step_per_year = t_step_per_year
        assert maturity is not None or (start_date is not None and end_date is not None), "Error: 到期时间或起止时间必须输入一个"
        self.start_date = start_date if start_date is not None else global_evaluation_date()  # 起始时间
        self.end_date = end_date if end_date is not None else (
                self.start_date + datetime.timedelta(days=round(maturity * annual_days.value)))  # 结束时间
        self.maturity = maturity if maturity is not None else (self.end_date - self.start_date).days / annual_days.value

        self.underlying_list = []  # 标的资产列表
        self.vanilla_list = []  # 期权列表
        if s is not None and r is not None and q is not None and vol is not None:
            self.engine = AnalyticVanillaEuEngine(s=s, r=r, q=q, vol=vol)
            self.default_engine = True
        else:
            self.engine = None
            self.default_engine = False

    def add_underlying(self, s0, position=1.0):
        """添加标的资产
        Args:
            s0: float，标的资产期初价格
            position: float，标的资产头寸
        Returns:
        """
        current_underlying = UnderlyingAsset(s0)
        if self.default_engine:
            current_underlying.set_pricing_engine(self.engine)
        self.underlying_list.append((current_underlying, position))
        self.show_current_portfolio()

    def add_vanilla(self, strike, callput: CallPut, position, maturity=None, start_date=None, end_date=None):
        """添加香草期权, 仅支持欧式期权
        Args:
            strike: float, 行权价
            callput: 看涨看跌类型, CallPut枚举类
            position: float, 香草期权头寸
            maturity: float, 年化自然日期限
            start_date: datetime.date, 期权起始日期
            end_date: datetime.date, 期权到期日期
        Returns:
        """
        if maturity is not None and start_date is not None and end_date is not None:
            assert maturity == (end_date - start_date).days / self.annual_days.value, "Error: 到期时间与起止时间不匹配！"
        _maturity = maturity if maturity is not None else self.maturity
        _start_date = start_date if start_date is not None else self.start_date
        _end_date = end_date if end_date is not None else self.end_date
        current_vanilla = VanillaOption(strike=strike, maturity=_maturity, start_date=_start_date, end_date=_end_date,
                                        callput=callput, exercise_type=ExerciseType.European)
        if self.default_engine:
            current_vanilla.set_pricing_engine(self.engine)
        self.vanilla_list.append((current_vanilla, position))
        self.show_current_portfolio()

    def __str__(self):
        return "自定义香草组合"

    def __repr__(self):
        """返回期权的描述"""
        description = f"{str(self)} - 当前资产组合为：\n"
        for vanilla, position in self.vanilla_list:
            description += f"行权价为{vanilla.strike}的{vanilla.callput.name}, 期限为{vanilla.maturity}，头寸为：{position}\n"
        return description

    def show_current_portfolio(self):
        """展示当前组合的标的资产和香草期权信息"""
        logging.info(repr(self))

    def del_underlying(self, index):
        """删除第index个标的资产"""
        logging.info(f"删除第{index}个标的资产")
        del self.underlying_list[index - 1]
        self.show_current_portfolio()

    def del_vanilla(self, index):
        """删除第index个香草期权"""
        logging.info(f"删除第{index}个香草期权")
        del self.vanilla_list[index - 1]
        self.show_current_portfolio()

    def set_pricing_engine(self, engine):
        """设置定价引擎"""
        self.default_engine = False
        self.engine = engine
        process = self.engine.process
        logging.info(f"{self}当前定价方法为{engine.engine_type.value}")
        for underlying, position in self.underlying_list:
            underlying.set_pricing_engine(engine)
        if engine.engine_type == EngineType.AnEngine:
            engine = AnalyticVanillaEuEngine(process)
            for vanilla, position in self.vanilla_list:
                vanilla.set_pricing_engine(engine)

    @time_this
    def price(self, t: datetime.date = None, spot=None):
        """计算期权价格
        Args:
            t: datetime.date，计算期权价格的日期
            spot: float，标的价格
        Returns: 期权现值
        """
        self.validate_parameters(t=t)
        result = 0
        for asset, position in self.underlying_list:
            result += position * asset.price(t=t, spot=spot)
        if self.engine.engine_type in [EngineType.AnEngine, EngineType.PdeEngine]:
            for vanilla, position in self.vanilla_list:
                result += position * vanilla.price(t=t, spot=spot)
        else:
            result += self.engine.calc_present_value(prod=self, t=t, spot=spot)
        return result


class Spread(VanillaPortfolio):
    """价差期权"""

    def __init__(self, lower_strike, upper_strike, callput: CallPut, maturity=None, start_date=None, end_date=None,
                 trade_calendar=CN_CALENDAR, annual_days=AnnualDays.N365, t_step_per_year=243,
                 s=None, r=None, q=None, vol=None):
        super().__init__(maturity=maturity, start_date=start_date, end_date=end_date, trade_calendar=trade_calendar,
                         annual_days=annual_days, t_step_per_year=t_step_per_year, s=s, r=r, q=q, vol=vol)
        self.lower_strike = lower_strike
        self.upper_strike = upper_strike
        self.callput = callput
        self.add_vanilla(strike=self.lower_strike, start_date=self.start_date, end_date=self.end_date,
                         callput=CallPut.Call, position=self.callput.value)
        self.add_vanilla(strike=self.upper_strike, start_date=self.start_date, end_date=self.end_date,
                         callput=CallPut.Call, position=-self.callput.value)

    def __str__(self):
        if self.callput == CallPut.Call:
            return r"牛市价差期权"
        return r"熊市价差期权"


class Butterfly(VanillaPortfolio):
    """蝶式期权"""

    def __init__(self, lower_strike, middle_strike, upper_strike, callput: CallPut, maturity=None, start_date=None,
                 end_date=None, trade_calendar=CN_CALENDAR, annual_days=AnnualDays.N365, t_step_per_year=243,
                 s=None, r=None, q=None, vol=None):
        super().__init__(maturity=maturity, start_date=start_date, end_date=end_date, trade_calendar=trade_calendar,
                         annual_days=annual_days, t_step_per_year=t_step_per_year, s=s, r=r, q=q, vol=vol)
        self.lower_strike = lower_strike
        self.middle_strike = middle_strike
        self.upper_strike = upper_strike
        self.callput = callput
        self.add_vanilla(strike=self.lower_strike, start_date=self.start_date, end_date=self.end_date,
                         callput=CallPut.Call, position=self.callput.value)
        self.add_vanilla(strike=self.middle_strike, start_date=self.start_date, end_date=self.end_date,
                         callput=CallPut.Call, position=-2 * self.callput.value)
        self.add_vanilla(strike=self.upper_strike, start_date=self.start_date, end_date=self.end_date,
                         callput=CallPut.Call, position=self.callput.value)

    def __str__(self):
        if self.callput == CallPut.Call:
            return r"^蝶式期权"
        return r"∨蝶式期权"


class Strangle(VanillaPortfolio):
    """宽跨式"""

    def __init__(self, lower_strike, upper_strike, callput: CallPut, maturity=None, start_date=None, end_date=None,
                 trade_calendar=CN_CALENDAR, annual_days=AnnualDays.N365, t_step_per_year=243,
                 s=None, r=None, q=None, vol=None):
        super().__init__(maturity=maturity, start_date=start_date, end_date=end_date, trade_calendar=trade_calendar,
                         annual_days=annual_days, t_step_per_year=t_step_per_year, s=s, r=r, q=q, vol=vol)
        self.lower_strike = lower_strike
        self.upper_strike = upper_strike
        self.callput = callput
        self.add_vanilla(strike=self.lower_strike, start_date=self.start_date, end_date=self.end_date,
                         callput=CallPut.Call, position=self.callput.value)
        self.add_vanilla(strike=self.upper_strike, start_date=self.start_date, end_date=self.end_date,
                         callput=CallPut.Call, position=self.callput.value)

    def __str__(self):
        if self.callput == CallPut.Call:
            return r"v宽跨式"
        return r"^宽跨式"


class Straddle(VanillaPortfolio):
    """跨式"""

    def __init__(self, strike, callput: CallPut, maturity=None, start_date=None, end_date=None,
                 trade_calendar=CN_CALENDAR, annual_days=AnnualDays.N365, t_step_per_year=243,
                 s=None, r=None, q=None, vol=None):
        super().__init__(maturity=maturity, start_date=start_date, end_date=end_date, trade_calendar=trade_calendar,
                         annual_days=annual_days, t_step_per_year=t_step_per_year, s=s, r=r, q=q, vol=vol)
        self.strike = strike
        self.callput = callput
        self.add_vanilla(strike=self.strike, start_date=self.start_date, end_date=self.end_date, callput=CallPut.Put,
                         position=self.callput.value)
        self.add_vanilla(strike=self.strike, start_date=self.start_date, end_date=self.end_date, callput=CallPut.Call,
                         position=self.callput.value)

    def __str__(self):
        if self.callput == CallPut.Call:
            return r"v跨式"
        return r"^跨式"


class Condor(VanillaPortfolio):
    """鹰式"""

    def __init__(self, strike1, strike2, strike3, strike4, callput: CallPut, maturity=None, start_date=None,
                 end_date=None, trade_calendar=CN_CALENDAR, annual_days=AnnualDays.N365, t_step_per_year=243,
                 s=None, r=None, q=None, vol=None):
        super().__init__(maturity=maturity, start_date=start_date, end_date=end_date, trade_calendar=trade_calendar,
                         annual_days=annual_days, t_step_per_year=t_step_per_year, s=s, r=r, q=q, vol=vol)
        self.strike1 = strike1
        self.strike2 = strike2
        self.strike3 = strike3
        self.strike4 = strike4
        self.callput = callput
        self.add_vanilla(strike=self.strike1, start_date=self.start_date, end_date=self.end_date, callput=CallPut.Call,
                         position=self.callput.value)
        self.add_vanilla(strike=self.strike2, start_date=self.start_date, end_date=self.end_date, callput=CallPut.Call,
                         position=-self.callput.value)
        self.add_vanilla(strike=self.strike3, start_date=self.start_date, end_date=self.end_date, callput=CallPut.Call,
                         position=-self.callput.value)
        self.add_vanilla(strike=self.strike4, start_date=self.start_date, end_date=self.end_date, callput=CallPut.Call,
                         position=self.callput.value)

    def __str__(self):
        if self.callput == CallPut.Call:
            return r"^鹰式"
        return r"v鹰式"


class DiscountCall(VanillaPortfolio):
    """折价看涨期权"""

    def __init__(self, strike, participate, callput=CallPut.Call, lower_strike=None, maturity=None, start_date=None,
                 end_date=None, trade_calendar=CN_CALENDAR, annual_days=AnnualDays.N365, t_step_per_year=243,
                 s=None, r=None, q=None, vol=None):
        super().__init__(maturity=maturity, start_date=start_date, end_date=end_date, trade_calendar=trade_calendar,
                         annual_days=annual_days, t_step_per_year=t_step_per_year, s=s, r=r, q=q, vol=vol)
        self.strike = strike
        if lower_strike is not None:
            self.lower_strike = lower_strike
        else:
            if callput == CallPut.Call:
                self.lower_strike = strike * participate
            elif callput == CallPut.Put:
                raise NotImplementedError("折价看跌期权不支持")
        self.participate = participate
        self.callput = callput
        self.add_vanilla(strike=self.lower_strike, start_date=self.start_date, end_date=self.end_date,
                         callput=CallPut.Call, position=self.callput.value)
        self.add_vanilla(strike=self.strike, start_date=self.start_date, end_date=self.end_date, callput=CallPut.Call,
                         position=-self.participate * self.callput.value)

    def __str__(self):
        return r"折价看涨"


class Collar(VanillaPortfolio):
    """领式"""

    def __init__(self, lower_strike, higher_strike, callput: CallPut, maturity=None, start_date=None, end_date=None,
                 trade_calendar=CN_CALENDAR, annual_days=AnnualDays.N365, t_step_per_year=243,
                 s=None, r=None, q=None, vol=None):
        super().__init__(maturity=maturity, start_date=start_date, end_date=end_date, trade_calendar=trade_calendar,
                         annual_days=annual_days, t_step_per_year=t_step_per_year, s=s, r=r, q=q, vol=vol)
        self.lower_strike = lower_strike
        self.higher_strike = higher_strike
        self.callput = callput
        self.add_underlying((self.lower_strike + self.higher_strike) / 2, 1)
        self.add_vanilla(strike=self.lower_strike, start_date=self.start_date, end_date=self.end_date,
                         callput=CallPut.Call, position=self.callput.value)
        self.add_vanilla(strike=self.higher_strike, start_date=self.start_date, end_date=self.end_date,
                         callput=CallPut.Call, position=-self.callput.value)

    def __str__(self):
        if self.callput == CallPut.Call:
            return r"买入领式"
        return r"卖出领式"


class RiskReversal(VanillaPortfolio):
    """风险反转"""

    def __init__(self, lower_strike, higher_strike, callput: CallPut, maturity=None, start_date=None, end_date=None,
                 trade_calendar=CN_CALENDAR, annual_days=AnnualDays.N365, t_step_per_year=243,
                 s=None, r=None, q=None, vol=None):
        super().__init__(maturity=maturity, start_date=start_date, end_date=end_date, trade_calendar=trade_calendar,
                         annual_days=annual_days, t_step_per_year=t_step_per_year, s=s, r=r, q=q, vol=vol)
        self.lower_strike = lower_strike
        self.higher_strike = higher_strike
        self.callput = callput
        self.add_vanilla(strike=self.lower_strike, start_date=self.start_date, end_date=self.end_date,
                         callput=CallPut.Call, position=-self.callput.value)
        self.add_vanilla(strike=self.higher_strike, start_date=self.start_date, end_date=self.end_date,
                         callput=CallPut.Call, position=self.callput.value)

    def __str__(self):
        if self.callput == CallPut.Call:
            return r"/风险反转"
        return r"\风险反转"



#!/user/bin/env python
# -*- coding: utf-8 -*-
"""
Copyright (C) 2024 Galaxy Technologies
Licensed under the Apache License, Version 2.0
"""
from pricelib.common.utilities.enums import CallPut, InOut, UpDown, ExerciseType, PaymentType
from pricelib.common.processes import StochProcessBase
from pricelib.common.pricing_engine_base import AnalyticEngine


class CashFlowEngine(AnalyticEngine):
    """现金流闭式解定价引擎
    主要用于计算票息收益等现金流的现值"""
    updown = UpDown.Down
    inout = InOut.In
    callput = CallPut.Call
    payment_type = PaymentType.Expire
    exercise_type = ExerciseType.American

    def __init__(self, stoch_process: StochProcessBase = None, *, s=None, r=None, q=0, vol=0):
        """
        初始化现金流闭式解定价引擎
        Args:
            stoch_process: StochProcessBase随机过程对象
        在未设置stoch_process时，(stoch_process=None)，会默认创建BSMprocess，需要输入以下变量进行初始化
            s: float，标的价格
            r: float，无风险利率
            q: float，分红/融券率
            vol: float，波动率
        """
        super().__init__(stoch_process=stoch_process, s=s, r=r, q=q, vol=vol)
        self.payment_date = None
        self.cashflow = None

    # pylint: disable=arguments-differ
    def calc_present_value(self, prod):
        """
        计算现金流现值
        Args:
            prod: Product产品对象
        Returns:
            pv: 现金流现值
        """
        pv = prod.cashflow * self.process.interest.disc_factor(t2=prod.payment_date)
        return pv



#!/user/bin/env python
# -*- coding: utf-8 -*-
"""
Copyright (C) 2024 Galaxy Technologies
Licensed under the Apache License, Version 2.0
"""
import datetime
from pricelib.common.utilities.enums import CallPut, ExerciseType
from pricelib.common.time import CN_CALENDAR, AnnualDays, global_evaluation_date
from pricelib.common.utilities.utility import time_this, logging
from pricelib.common.product_base.option_base import OptionBase
from pricelib.pricing_engines.analytic_engines import AnalyticVanillaEuEngine, AnalyticVanillaAmEngine


class VanillaOption(OptionBase):
    """香草期权，包括欧式期权和美式期权
    均支持解析解、MC、PDE、Quad、二叉树"""

    def __init__(self, strike, callput=CallPut.Call, exercise_type=ExerciseType.European, *, engine=None,
                 maturity=None, start_date=None, end_date=None, trade_calendar=CN_CALENDAR, annual_days=AnnualDays.N365,
                 t_step_per_year=243, s=None, r=None, q=None, vol=None):
        """构造函数
        产品参数:
            strike: float，行权价
            callput: 看涨看跌类型，CallPut枚举类，Call/Put
            exercise_type: 行权方式，ExerciseType枚举类，European/American
            engine: 定价引擎，PricingEngine类
                    解析解: 欧式AnalyticVanillaEuEngine/美式AnalyticVanillaAmEngine
                    蒙特卡洛: MCVanillaEngine 欧式/美式
                    PDE: FdmVanillaEngine 欧式/美式
                    积分法: QuadVanillaEngine 欧式/美式
                    二叉树: BiTreeVanillaEngine 欧式/美式
        时间参数: 要么输入年化期限，要么输入起始日和到期日
            maturity: float，年化期限
            start_date: datetime.date，起始日
            end_date: datetime.date，到期日
            trade_calendar: 交易日历，Calendar类，默认为中国内地交易日历
            annual_days: int，每年的自然日数量
            t_step_per_year: int，每年的交易日数量
        可选参数:
            若未提供引擎的情况下，提供了标的价格、无风险利率、分红/融券率、波动率，
            则默认使用解析解定价引擎 - (欧式期权BSM公式, 美式期权BAW近似解)
            s: float，标的价格
            r: float，无风险利率
            q: float，分红/融券率
            vol: float，波动率
        """
        super().__init__()
        self.trade_calendar = trade_calendar
        self.annual_days = annual_days
        assert maturity is not None or (start_date is not None and end_date is not None), "Error: 到期时间或起止时间必须输入一个"
        self.start_date = start_date if start_date is not None else global_evaluation_date()  # 起始时间
        self.end_date = end_date if end_date is not None else (
                self.start_date + datetime.timedelta(days=round(maturity * annual_days.value)))  # 结束时间
        self.maturity = maturity if maturity is not None else (self.end_date - self.start_date).days / annual_days.value
        self.strike = strike
        self.callput = callput
        self.exercise_type = exercise_type
        self.t_step_per_year = t_step_per_year
        if engine is not None:
            self.set_pricing_engine(engine)
        elif s is not None and r is not None and q is not None and vol is not None:
            if exercise_type == ExerciseType.European:
                default_engine = AnalyticVanillaEuEngine(s=s, r=r, q=q, vol=vol)
                self.set_pricing_engine(default_engine)
            elif exercise_type == ExerciseType.American:
                default_engine = AnalyticVanillaAmEngine(s=s, r=r, q=q, vol=vol, an_method="BAW")
                self.set_pricing_engine(default_engine)

    def set_pricing_engine(self, engine):
        """切换定价引擎"""
        self.engine = engine
        logging.info(f"{self}当前定价方法为{engine.engine_type.value}")

    def __repr__(self):
        """返回期权的描述"""
        return f"{self.exercise_type.value}{self.callput.name}香草期权"

    @time_this
    def price(self, t: datetime.date = None, spot=None):
        """计算期权价格
        Args:
            t: datetime.date，计算期权价格的日期
            spot: float，标的价格
        Returns: 期权现值
        """
        self.validate_parameters(t=t)
        price = self.engine.calc_present_value(prod=self, t=t, spot=spot)
        return price



#!/user/bin/env python
# -*- coding: utf-8 -*-
"""
Copyright (C) 2024 Galaxy Technologies
Licensed under the Apache License, Version 2.0
"""
import numpy as np
from numba import njit, prange
from scipy.interpolate import interp1d, UnivariateSpline


# pylint: disable=invalid-name
def TDMA(X, d):
    """三对角矩阵求解，Thomas算法，the Tridiagonal Matrix Algorithm，是高斯消元法对三对角矩阵的特例。
    Args:
        X: np.adarray，三对角系数矩阵
        d: np.adarray，方程右端向量
    Returns:
        p: np.adarray，方程的解向量，Xp=d
    """
    n = len(d)
    a = np.array(X[np.arange(n - 1) + 1, np.arange(n - 1)])[0]
    b = np.array(X[np.arange(n), np.arange(n)])[0]
    c = np.array(X[np.arange(n - 1), np.arange(n - 1) + 1])[0]
    w = np.zeros(n - 1, float)
    g = np.zeros(n, float)
    p = np.zeros(n, float)

    w[0] = c[0] / b[0]
    g[0] = d[0] / b[0]

    for i in range(1, n - 1):
        w[i] = c[i] / (b[i] - a[i - 1] * w[i - 1])
    for i in range(1, n):
        g[i] = (d[i] - a[i - 1] * g[i - 1]) / (b[i] - a[i - 1] * w[i - 1])
    p[n - 1] = g[n - 1]
    for i in range(n - 1, 0, -1):
        p[i - 1] = g[i - 1] - w[i - 1] * p[i]
    return p.reshape(n, )


# pylint: disable=invalid-name
@njit(cache=True, fastmath=True)  # 不要使用parallel=True，因为计算有顺序，不能并行。否则计算结果不对
def TDMA_jit(X, d):
    """使用numba.jit加速的三对角矩阵求解Thomas算法，the Tridiagonal Matrix Algorithm
    Args:
        X: np.adarray，三对角系数矩阵
        d: np.adarray，方程右端向量
    Returns:
        p: np.adarray，方程的解向量，Xp=d
    """
    # 将矩阵的三条对角线取出来
    n = len(d)
    a = np.zeros(n - 1)
    b = np.zeros(n)
    c = np.zeros(n - 1)
    for i in prange(n):
        if i > 0:
            a[i - 1] = X[i, i - 1]
        b[i] = X[i, i]
        if i < n - 1:
            c[i] = X[i, i + 1]
    # Thomas算法
    for i in prange(1, n):
        m = a[i - 1] / b[i - 1]
        b[i] = b[i] - m * c[i - 1]
        d[i] = d[i] - m * d[i - 1]
    p = b
    p[-1] = d[-1] / b[-1]
    indices = np.arange(n - 2, -1, -1)  # 创建逆序索引数组
    for j in prange(n - 1):
        i = indices[j]  # 使用逆序索引
        p[i] = (d[i] - c[i] * p[i + 1]) / b[i]
    return p


# pylint: disable=invalid-name
@njit(cache=True, fastmath=True)  # 不要使用parallel=True，因为计算有顺序，不能并行。否则计算结果不对
def TDMA_ldu_jit(a, b, c, d):
    """使用numba.jit大幅加速的三对角矩阵求解Thomas算法，the Tridiagonal Matrix Algorithm
    Args:
        a: np.adarray，三对角系数矩阵的下对角线元素向量
        b: np.adarray，三对角系数矩阵的主对角线元素向量
        c: np.adarray，三对角系数矩阵的上对角线元素向量
        d: np.adarray，方程右端向量
    Returns:
        p: np.adarray，方程的解向量，Xp=d
    """
    n = len(d)
    for i in prange(1, n):
        m = a[i - 1] / b[i - 1]
        b[i] = b[i] - m * c[i - 1]
        d[i] = d[i] - m * d[i - 1]
    p = b
    p[-1] = d[-1] / b[-1]
    indices = np.arange(n - 2, -1, -1)  # 创建逆序索引数组
    for j in prange(n - 1):
        i = indices[j]  # 使用逆序索引
        p[i] = (d[i] - c[i] * p[i + 1]) / b[i]
    return p


class LinearFlat(interp1d):
    """自定义以最边缘点的值水平外推的线性插值，两侧外推时，使用边界值y[0], y[-1]"""

    def __init__(self, x, y):
        """初始化，执行单变量线性插值interp1d"""
        super().__init__(x, y, kind='linear', bounds_error=False, fill_value=(y[0], y[-1]))


class CubicSplineFlat(UnivariateSpline):
    """自定义以最边缘点的值水平外推的三次样条插值，
    s是平滑系数，是非负实数，用于控制拟合曲线的平滑程度，s越大平滑程度越大，s越小拟合曲线越接近原始数据点。当s=0时，曲线会穿过所有的原始数据点。
    ext控制x在外推时的行为，ext=3 or 'const', 以边界值水平外推。
    k控制平滑样条的程度，1~5的整数。默认k=3， 三次样条"""

    def __init__(self, x, y):
        """初始化，执行单变量插值UnivariateSpline"""
        super().__init__(x, y, s=0, ext=3)  #


class FlatCubicSpline(UnivariateSpline):
    """自定义边界上的一阶导和二阶导都是0的三次样条插值(这样边界值自然是水平外推的。)
    s是平滑系数，是非负实数，用于控制拟合曲线的平滑程度，s越大平滑程度越大，s越小拟合曲线越接近原始数据点。当s=0时，曲线会穿过所有的原始数据点。
    ext控制x在外推时的行为，ext=3 or 'const', 以边界值水平外推。
    k控制平滑样条的程度，1~5的整数。默认k=3， 三次样条"""

    def __init__(self, x, y):
        """初始化，执行单变量插值UnivariateSpline
        x必须是升序的ascending
        为了模拟边界上的一阶导和二阶导都是0，将x的两端分别外扩一个小量epsilon，对应的y值与边界y值相同"""
        epsilon = 1e-8
        x = np.hstack((x[0] * (1 - epsilon), x, x[-1] * (1 + epsilon)))
        y = np.hstack((y[0], y, y[-1]))
        super().__init__(x, y, s=0, ext=3)



#!/user/bin/env python
# -*- coding: utf-8 -*-
"""
Copyright (C) 2024 Galaxy Technologies
Licensed under the Apache License, Version 2.0
"""
import numpy as np
from scipy.stats import norm
from pricelib.common.utilities.enums import CallPut, ExerciseType, PaymentType
from pricelib.common.time import global_evaluation_date
from pricelib.common.pricing_engine_base import AnalyticEngine


class AnalyticCashOrNothingEngine(AnalyticEngine):
    """二元(数字)期权-现金或无(cash or nothing)-闭式解定价引擎
    Reiner and Rubinstein(1991b) 支持欧式二元(现金或无)、美式二元(立即支付)、一触即付(到期支付)，
    支持M. Broadie, P. Glasserman, S.G. Kou(1999)离散观察调整"""

    # pylint: disable=invalid-name
    def calc_present_value(self, prod, t=None, spot=None):
        """计算现值
        Args:
            prod: Product产品对象
            t: datetime.date，估值日; 如果是None，则使用全局估值日globalEvaluationDate
            spot: float，估值日标的价格，如果是None，则使用随机过程的当前价格
        Returns: float，现值
        """
        calculate_date = global_evaluation_date() if t is None else t
        tau = prod.trade_calendar.business_days_between(calculate_date, prod.end_date) / prod.t_step_per_year
        if spot is None:
            spot = self.process.spot()
        r = self.process.interest(tau)
        q = self.process.div(tau)
        vol = self.process.vol(tau, spot)

        v = r - q - vol ** 2 / 2
        mu = (r - q) / vol ** 2 - 0.5
        K_S = prod.strike / spot
        denominator = vol * np.sqrt(tau)
        # 欧式
        if prod.exercise_type == ExerciseType.European:
            return prod.rebate * np.exp(-r * tau) * norm.cdf(prod.callput.value * (v * tau - np.log(K_S)) / denominator)
        if prod.exercise_type == ExerciseType.American:
            # 离散观察期权，M. Broadie, P. Glasserman, S.G. Kou(1999) 在连续观察期权解析解上加调整项
            if prod.discrete_obs_interval is not None:
                if prod.callput == CallPut.Call:  # 向上
                    strike = prod.strike * np.exp(0.5826 * vol * np.sqrt(prod.discrete_obs_interval))
                elif prod.callput == CallPut.Put:  # 向下
                    strike = prod.strike * np.exp(-0.5826 * vol * np.sqrt(prod.discrete_obs_interval))
                else:
                    raise ValueError("不支持的看涨看跌类型")
            else:
                strike = prod.strike
            K_S = strike / spot
            # 美式，立即支付
            if prod.payment_type == PaymentType.Hit:
                lambda_ = np.sqrt(mu ** 2 + 2 * r / vol ** 2)
                return prod.rebate * (K_S ** (mu + lambda_) * norm.cdf(
                    -prod.callput.value * (np.log(K_S) / denominator + lambda_ * denominator)) +
                                      K_S ** (mu - lambda_) * norm.cdf(
                            -prod.callput.value * (np.log(K_S) / denominator - lambda_ * denominator)))
            # 美式，到期支付
            if prod.payment_type == PaymentType.Expire:
                return prod.rebate * np.exp(-r * tau) * (
                        norm.cdf((v * tau - np.log(K_S)) / denominator * prod.callput.value) +
                        K_S ** (2 * mu) * norm.cdf(
                            -(v * tau + np.log(K_S)) / denominator * prod.callput.value))
            raise ValueError("不支持的Payment type")



#!/user/bin/env python
# -*- coding: utf-8 -*-
"""
Copyright (C) 2024 Galaxy Technologies
Licensed under the Apache License, Version 2.0
"""
import numpy as np
from pricelib.common.utilities.enums import CallPut
from pricelib.common.pricing_engine_base import McEngine
from pricelib.common.time import global_evaluation_date


class MCAutoCallEngine(McEngine):
    """AutoCall Note(小雪球) MonteCarlo 模拟定价引擎"""

    def calc_present_value(self, prod, t=None, spot=None):
        """计算现值
        Args:
            prod: Product产品对象
            t: datetime.date，估值日; 如果是None，则使用全局估值日globalEvaluationDate
            spot: float，估值日标的价格，如果是None，则使用随机过程的当前价格
        Returns: float，现值
        """
        calculate_date = global_evaluation_date() if t is None else t
        _maturity = (prod.end_date - calculate_date).days / prod.annual_days.value
        _maturity_business_days = prod.trade_calendar.business_days_between(calculate_date, prod.end_date)
        obs_dates = prod.obs_dates.count_business_days(calculate_date)
        obs_dates = np.array([num for num in obs_dates if num >= 0])
        # 支付日 - 计算起始日到支付日的天数，用于计算应支付的敲出收益
        calculate_start_diff = (calculate_date - prod.start_date).days
        pay_dates = prod.pay_dates.count_calendar_days(prod.start_date)
        pay_dates = np.array([num / prod.annual_days.value for num in pay_dates if num >= calculate_start_diff])
        assert len(obs_dates) == len(pay_dates), f"Error: {prod}的观察日和付息日长度不一致"
        # 支付日 - 计算估值日到支付日的天数，用于折现
        pay_dates_tau = prod.pay_dates.count_calendar_days(calculate_date)
        pay_dates_tau = np.array([num / prod.annual_days.value for num in pay_dates_tau if num >= 0])
        # 经过估值日截断列表，例如prod.barrier_out有22个，存续一年时估值，_barrier_out只有12个
        _barrier_out = prod.barrier_out[-len(obs_dates):].copy()
        _coupon_out = prod.coupon_out[-len(obs_dates):].copy()
        if spot is None:
            spot = self.process.spot()
        else:
            self.reset_paths_flag()  # 重置路径标志位，重新生成路径

        s_paths = self.path_generator(n_step=_maturity_business_days, spot=spot,
                                      t_step_per_year=prod.t_step_per_year).copy()

        barrier = np.tile(_barrier_out, (self.n_path, 1)).T
        # 记录每条路径的具体敲出日，如果无敲出则保留inf

        knock_out_scenario = np.tile(obs_dates, (self.n_path, 1)).T

        if prod.callput == CallPut.Call:
            knock_out_scenario = np.where(s_paths[obs_dates] >= barrier, knock_out_scenario, np.inf)
        elif prod.callput == CallPut.Put:
            knock_out_scenario = np.where(s_paths[obs_dates] <= barrier, knock_out_scenario, np.inf)
        knock_out_date = np.min(knock_out_scenario, axis=0)
        is_knock_out = knock_out_date != np.inf
        # 红利票息
        hold_payoff = ((self.n_path - np.sum(is_knock_out)) * prod.s0 * (prod.coupon_div * _maturity
                       + prod.margin_lvl) * self.process.interest.disc_factor(_maturity))
        # 敲出部分
        # 敲出对应计息时长
        obs_to_pay = dict(zip(obs_dates, pay_dates))
        pay_dates_vec = np.vectorize(obs_to_pay.get)(knock_out_date[is_knock_out])
        # 敲出对应折现时长
        obs_to_pay_tau = dict(zip(obs_dates, pay_dates_tau))
        pay_dates_tau_vec = np.vectorize(obs_to_pay_tau.get)(knock_out_date[is_knock_out])
        # 敲出对应票息
        obs_to_coupon = dict(zip(obs_dates, _coupon_out))
        coupon_out = np.vectorize(obs_to_coupon.get)(knock_out_date[is_knock_out])
        # 敲出部分的收益
        knock_out_payoff = np.sum(prod.s0 * (coupon_out * pay_dates_vec + prod.margin_lvl)
                                  * self.process.interest.disc_factor(pay_dates_tau_vec))
        value = (hold_payoff + knock_out_payoff) / self.n_path
        return value



#!/user/bin/env python
# -*- coding: utf-8 -*-
"""
Copyright (C) 2024 Galaxy Technologies
Licensed under the Apache License, Version 2.0
"""
import numpy as np
from pricelib.common.utilities.enums import StatusType
from pricelib.common.time import CN_CALENDAR, AnnualDays
from .autocallable_base import AutocallableBase


class BothDownSnowball(AutocallableBase):
    """双降雪球结构产品类"""

    def __init__(self, s0, barrier_out_start, barrier_out_step, barrier_in, coupon_ko_start, coupon_ko_step,
                 coupon_div=None, lock_term=3, maturity=None, start_date=None, end_date=None,
                 obs_dates=None, pay_dates=None, parti_in=1, margin_lvl=1, status=StatusType.NoTouch, engine=None,
                 trade_calendar=CN_CALENDAR, annual_days=AnnualDays.N365, t_step_per_year=243,
                 s=None, r=None, q=None, vol=None):
        """继承自动赎回基类AutocallableBase的参数，详见AutocallableBase的__init__方法
        Args: 以下为区别于经典雪球的参数
            barrier_out_start: float, 敲出障碍价起始值，绝对值/百分比
            barrier_out_step: float, 降敲步长，绝对值/百分比
            lock_term: int，锁定期，单位为月，锁定期内不触发敲出
            coupon_ko_start: float, 敲出票息起始值，百分比，年化
            coupon_ko_step: float, 降票息步长，百分比，年化
        """
        super().__init__(s0=s0, maturity=maturity, start_date=start_date, end_date=end_date, lock_term=lock_term,
                         trade_calendar=trade_calendar, obs_dates=obs_dates, pay_dates=pay_dates, status=status,
                         annual_days=annual_days, parti_in=parti_in, margin_lvl=margin_lvl,
                         t_step_per_year=t_step_per_year, engine=engine, s=s, r=r, q=q, vol=vol)
        len_obs_dates = len(self.obs_dates.date_schedule)

        barrier_out = np.arange(barrier_out_start, barrier_out_start - barrier_out_step * len_obs_dates,
                                -barrier_out_step)[:len_obs_dates]
        self.barrier_out = barrier_out
        self.barrier_in = np.ones(len_obs_dates) * barrier_in
        coupon_out = np.arange(coupon_ko_start, coupon_ko_start - coupon_ko_step * len_obs_dates, -coupon_ko_step)[
                     :len_obs_dates]
        self.coupon_out = coupon_out
        self.coupon_div = coupon_div if coupon_div is not None else coupon_out[-1]
        self.parti_out = 0
        self.strike_upper = s0
        self.strike_lower = 0

    def __repr__(self):
        """返回期权的描述"""
        return "双降雪球"



#!/user/bin/env python
# -*- coding: utf-8 -*-
"""
Copyright (C) 2024 Galaxy Technologies
Licensed under the Apache License, Version 2.0
"""
import warnings
from abc import ABCMeta
from functools import lru_cache
import numpy as np
import numba as nb
from scipy.stats import norm
from ..processes import StochProcessBase
from ..utilities.enums import QuadMethod, ProcessType, EngineType
from ..utilities.patterns import HashableArray
from .engine_base import PricingEngineBase


@lru_cache(maxsize=1)
@nb.njit(cache=True, fastmath=True)
def get_quad_vector(n_points, quad_method):
    """生成积分权重向量
    Args:
        n_points: int，积分点个数
        quad_method: 数值积分方式，QuadMethod枚举类，Trapezoid梯形法则/Simpson辛普森法则
    Returns:
        quad_vector: np.ndarray，积分权重向量
        quad_index: int，数值积分方法对应的调整系数
    """
    quad_vector = np.ones(n_points)
    if quad_method == QuadMethod.Trapezoid:
        quad_vector[1:-1] *= 2
        quad_index = 2
    else:  # quad_method == QuadMethod.Simpson:
        quad_vector[1::2] *= 4
        quad_vector[0::2] *= 2
        quad_vector[0] = quad_vector[-1] = 1
        quad_index = 3
    return quad_vector, quad_index


@nb.njit(cache=True, fastmath=True, parallel=True)
def step_backward_jit(x, y, v, t, r, q, vol, quad_vector, quad_index):
    """向前递推一步，更新期权价值向量，使用jit加速
    Args:
        x: np.ndarray，需要计算的未知时点，对应的标的价格向量的绝对值，不是对数价格
        y: np.ndarray，已知时点，对应的标的价格向量的绝对值，不是对数价格
        v: np.ndarray，已知时点，对应的期权价值向量
        t: float，时间间隔
        r: float，无风险利率
        q: float，分红率
        vol: float，波动率
        quad_vector: np.ndarray，积分权重向量
        quad_index: int，数值积分方法对应的调整系数
    Returns:
        target_v: np.ndarray(HashableArray)，更新后的期权价值向量
    """
    upper_barrier = np.max(y)
    lower_barrier = np.min(y)

    tvar = 0.5 * vol ** 2 * t
    rho = 1 / (2 * np.sqrt(np.pi * tvar) * y)
    rho = rho * np.exp(-1 / (4 * tvar) * (np.log(y / x.reshape(-1, 1)) - (r - q) * t + tvar) ** 2)
    target_v = np.dot(quad_vector, (rho * v).T) * (upper_barrier - lower_barrier) / (y.shape[0] - 1) / quad_index
    target_v *= np.exp(-r * t)
    return target_v


@nb.njit(cache=True, fastmath=True)
def fft_convolve_jit(v, pdf, t, r, ln_ds, quad_vector, quad_index):
    """快速傅里叶变换计算卷积，使用jit加速
    FFT加速多项式乘法的原理: 把空间域中复杂的卷积给转换到频率去进行乘法运算，然后再通过逆变换，就可以得到在空间域中的卷积运算的结果
    Args:
        v: np.ndarray，已知时点，对应的期权价值向量
        pdf: np.ndarray，未知时点到已知时点，对应的转移概率密度向量
        t: float，时间间隔
        r: float，无风险利率
        ln_ds: float, 积分变量，对数价格均匀格点的间隔
        quad_vector: np.ndarray，积分权重向量
        quad_index: int，数值积分方法对应的调整系数
    Returns:
        np.ndarray, 卷积结果
    """
    len_v = v.size
    len_pdf = pdf.size
    v = np.hstack((v * quad_vector, np.zeros(len_pdf - len_v)))  # 对v加倍次数界，乘以辛普森积分权重
    fft_v = np.fft.fft(v)  # 对一个方程的系数进行傅里叶变换
    fft_pdf = np.fft.fft(pdf)  # 对第二个方程的系数进行傅里叶变换
    fft_res = fft_v * fft_pdf  # 对两个FFT变换结果进行相乘 (numpy中已经实现了数组相乘)
    fft_res = np.fft.ifft(fft_res).real
    fft_res = fft_res[len_v - 1:] / quad_index * ln_ds * np.exp(- r * t)
    return fft_res


class QuadEngine(PricingEngineBase, metaclass=ABCMeta):
    """数值积分法定价引擎，目前只支持常数r、q、vol"""
    engine_type = EngineType.QuadEngine

    def __init__(self, stoch_process: StochProcessBase = None, quad_method=QuadMethod.Simpson, n_points=1301,
                 *, s=None, r=None, q=None, vol=None):
        """初始化数值积分法定价引擎
        Args:
            stoch_process: StochProcessBase，随机过程
            quad_method: 数值积分方法，QuadMethod枚举类，Trapezoid梯形法则/Simpson辛普森法则
            n_points: int，积分点个数 (需要注意，辛普森法则的n_points必须是奇数)
        在未设置stoch_process时，(stoch_process=None)，会默认创建BSMprocess，需要输入以下变量进行初始化
            s: float，标的价格
            r: float，无风险利率
            q: float，分红/融券率
            vol: float，波动率
        """
        super().__init__(stoch_process=stoch_process, s=s, r=r, q=q, vol=vol)
        self.n_points = n_points
        self.quad_method = quad_method
        # 以下为计算过程的中间变量
        self.backward_steps = 1
        self.r = None
        self.q = None
        self.vol = None
        # 积分法相关变量
        self.ln_s_vec = None
        self.ln_ds = None
        self.ln_boundary = None
        self.prod = None

    def set_stoch_process(self, stoch_process: StochProcessBase):
        assert stoch_process.process_type == ProcessType.BSProcess1D, 'Error: 数值积分方法只能使用1维BSM动态过程'
        self.process = stoch_process

    def _check_method_params(self):
        """n_points数检查：设置价格格点数下限，检查格点数是否是奇数"""
        if self.n_points % 2 == 0:
            self.n_points += 1

    def set_quad_params(self, r=None, q=None, vol=None):
        """设置常数r、q、vol"""
        self.r = r
        self.q = q
        self.vol = vol

    def init_grid(self, spot, vol, tau, barrier_low=None, barrier_high=None):
        """初始化积分法价格格点
        为了使用快速傅里叶变换进行卷积运算，需要将价格取对数，从而将分布转化为对称的正态分布
        如果存在高低障碍价格，需要确定价格格点的边界
        Args:
            spot: float, 当前标的价格
            vol: float, 波动率
            tau: float, 年化到期时间
            barrier_low: float, 存续期内，所有低障碍价格的最小值
            barrier_high: float, 存续期内，所有高障碍价格的最大值
        Returns:
            self.ln_s_vec: np.ndarray, 对数价格的均匀格点向量
            self.ln_ds: np.ndarray, 对数价格的均匀格点的间距
            self.ln_boundary: np.ndarray, 只有两个元素，(对数低价格边界，对数高价格边界)
        """
        c = np.exp(10 * vol * np.sqrt(tau) + 0.5 * vol ** 2 * tau)
        boundary = np.array([spot / c, spot * c])
        if barrier_low is not None:
            boundary[0] = np.max(barrier_low, boundary[0])
        if barrier_high is not None:
            boundary[1] = np.min(barrier_high, boundary[1])
        boundary = np.log(boundary)
        self.ln_s_vec = HashableArray(np.linspace(boundary[0], boundary[1], self.n_points))
        self.ln_ds = self.ln_s_vec[1] - self.ln_s_vec[0]
        self.ln_boundary = HashableArray(boundary)

    def Vma(self, s, km, epsilon, t):
        """ Vma = 欧式香草期权中的资产或无的价值
        例如看涨期权，到期时刻，St > strike, Vma = St; St < strike, Vma = 0
        Args:
            s: 期初价格
            km: 行权价
            epsilon: 1 or -1，看涨期权为1，看跌期权为-1
            t: 年化到期时间
        Returns: Vma = 欧式香草期权中的资产或无的部分
        """
        tau = 0.5 * self.vol ** 2 * t
        d1 = (np.log(s / km) + (self.r - self.q + 0.5 * self.vol ** 2) * t) / (self.vol * np.sqrt(t))
        return np.exp((-2 * self.q * tau) / (self.vol ** 2)) * s * norm.cdf(epsilon * d1)

    def Vmb(self, s, km, epsilon, t):
        """ Vmb = 欧式香草期权中的现金或无的价值
        例如看涨期权，到期时刻，St > strike, Vmb = 1 ; St < strike, Vmb = 0
        Args:
            s: 期初价格
            km: 行权价
            epsilon: 1 or -1，看涨期权为1，看跌期权为-1
            t: 年化到期时间
        Returns: Vmb = 欧式香草期权中的现金或无的部分
        """
        tau = 0.5 * self.vol ** 2 * t
        d1 = (np.log(s / km) + (self.r - self.q + 0.5 * self.vol ** 2) * t) / (self.vol * np.sqrt(t))
        d2 = d1 - self.vol * np.sqrt(t)
        return np.exp((-2 * self.r * tau) / (self.vol ** 2)) * norm.cdf(epsilon * d2)

    def quad_step_backward(self, x, y, v, t, quad_method=None):
        """直接计算数值积分: 向前递推一步，用vy推出vx，更新期权价值向量
        Args:
            x: np.ndarray，需要计算的未知时点，对应的标的价格向量的绝对值，不是对数价格
            y: np.ndarray，已知时点，对应的标的价格向量的绝对值，不是对数价格。为x之后的时点
            v: np.ndarray，已知时点，对应的期权价值向量vy
            t: float，时间间隔
            quad_method: 数值积分方式，QuadMethod枚举类，Trapezoid梯形法则/Simpson辛普森法则
        Returns:
            result: np.ndarray，更新后的期权价值向量vx
        """
        quad_method = self.quad_method if quad_method is None else quad_method
        quad_vector, quad_index = get_quad_vector(y.size, quad_method)
        result = step_backward_jit(x, y, v, t, self.r, self.q, self.vol, quad_vector, quad_index)
        result = np.array(result)
        return result

    def fft_step_backward(self, x, y, v, t):
        """FFT积分: 向前递推一步，用vy推出vx，更新期权价值向量
            当x, y使用相同间隔，并且分布具有对称性时，可以通过快速傅里叶变换加速计算
        Args:
            x: np.ndarray，需要计算的未知时点，对应的标的价格向量的对数值的均匀格点
            y: np.ndarray，已知时点，对应的标的价格向量的对数值的均匀格点。为x之后的时点
            v: np.ndarray，已知时点，对应的期权价值向量vy
            t: float，时间间隔
        Returns:
            result: np.ndarray，更新后的期权价值向量vx
        """
        quad_vector, quad_index = get_quad_vector(y.size, self.quad_method)
        result = fft_convolve_jit(v, self.get_pdf(x, y, t), self.r, t, self.ln_ds, quad_vector, quad_index)
        result = np.array(result)
        return result

    def get_pdf(self, x_vec, y_vec, t):
        """计算对数价格的概率密度函数，即x到y的转移概率密度 (仅FFT使用)
        Args:
            x_vec: np.ndarray, 未知时点的对数标的价格向量
            y_vec: np.ndarray, 已知时点的对数标的价格向量，为x之后的时点
            t: float, 时间间隔
        Returns:
            np.ndarray, x到y的转移概率密度
        """
        quad_vec = np.concatenate(((x_vec - x_vec[-1] + y_vec[0])[:-1], y_vec))
        quad_vec -= x_vec[0]
        pdf = norm.pdf(quad_vec, (self.r - self.q - 0.5 * self.vol ** 2) * t, self.vol * np.sqrt(t))
        return np.flip(pdf)

    @staticmethod
    @lru_cache
    def get_lns_vec(full_lns_vec, current_bound):
        """获取目标时刻的 需要FFT计算的价格向量 + [可能]低障碍调整价格向量 + [可能]高障碍调整价格向量
        Args:
            full_lns_vec: np.ndarray，初始生成的对数均匀价格格点向量
            current_bound: np.ndarray，障碍价格，只有两个元素，递推时目标时刻的[低障碍价格绝对值，高障碍价格绝对值]
        Returns:
            fft_lns_vec: np.ndarray, 需要FFT计算的价格向量
            lower_adj_vec: np.ndarray, 三个元素，对数的[低障碍B-，B-和L-的中点，价格向量最低点L-]; 若无需调整，为None
            upper_adj_vec: np.ndarray, 三个元素，对数的[价格向量最高点L+，L+和B+的中点，高障碍B+]; 若无需调整，为None
        """
        with warnings.catch_warnings():  # 临时忽略ln 0 = -inf的警告
            warnings.filterwarnings('ignore', category=RuntimeWarning)
            ln_bound = np.log(current_bound)
        fft_lns_vec = full_lns_vec[(full_lns_vec >= ln_bound[0]) & (full_lns_vec <= ln_bound[1])]
        lower_lns_vec = full_lns_vec[full_lns_vec < ln_bound[0]]
        upper_lns_vec = full_lns_vec[full_lns_vec > ln_bound[1]]
        if ln_bound[0] < fft_lns_vec[0] and not np.isinf(ln_bound[0]):
            lower_adj_vec = HashableArray([ln_bound[0], (ln_bound[0] + fft_lns_vec[0]) / 2, fft_lns_vec[0]])
        else:
            lower_adj_vec = None
        if ln_bound[1] > fft_lns_vec[0] and not np.isinf(ln_bound[1]):
            upper_adj_vec = HashableArray([fft_lns_vec[-1], (ln_bound[1] + fft_lns_vec[-1]) / 2, ln_bound[1]])
        else:
            upper_adj_vec = None
        return fft_lns_vec, lower_lns_vec, upper_lns_vec, lower_adj_vec, upper_adj_vec

    @staticmethod
    @lru_cache
    def exp_transfer(arg):
        return np.exp(arg)

    def fft_with_barrier(self, x, y, v, t, upper_adj_x=None, lower_adj_x=None,
                         upper_adj_y=None, lower_adj_y=None, upper_adj_v=None, lower_adj_v=None):
        """当存在障碍时的FFT积分，用vy推出vx
        当存在障碍时，障碍价格很可能不会恰好落在对数价格的均匀格点上，因此需要对最上/最下格点与障碍价格之间的区间单独积分，加入到FFT积分的结果中，
            上下边缘的积分固定使用辛普森法则.
        Args:
            x: np.ndarray，需要计算的未知时点，对应的标的价格向量的对数值的均匀格点(范围在高低障碍价之内的格点)
            y: np.ndarray，已知时点，对应的标的价格向量的对数值的均匀格点，为x之后的时点
            v: np.ndarray，已知时点，对应的期权价值向量vy
            t: float，时间间隔
            upper_adj_x: np.ndarray，未知时点，上边界3个对数价格点(最大格点，两者中点，对数高障碍价)
            lower_adj_x: np.ndarray，未知时点，下边界3个对数价格点(对数低障碍价，两者中点，最小格点)
            upper_adj_y: np.ndarray，已知时点，上边界3个对数价格点(最大格点，两者中点，对数高障碍价)
            lower_adj_y: np.ndarray，已知时点，下边界3个对数价格点(对数低障碍价，两者中点，最小格点)
            upper_adj_v: np.ndarray，已知时点，上边界3个对数价格点对应的期权价值向量
            lower_adj_v: np.ndarray，已知时点，下边界3个对数价格点对应的期权价值向量
        Returns:
            result: np.ndarray，更新后的期权价值向量vx
        """
        v_vec = self.fft_step_backward(x, y, v, t)
        # 由于quad_step_backward方法的输入参数是价格，而不是对数价格，此处需要做转换
        exp_x = self.exp_transfer(x)
        if lower_adj_y is not None:
            lower_exp_y = self.exp_transfer(lower_adj_y)
            v_vec += self.quad_step_backward(exp_x, lower_exp_y, lower_adj_v, t, quad_method=QuadMethod.Simpson)
        if upper_adj_y is not None:
            upper_exp_y = self.exp_transfer(upper_adj_y)
            v_vec += self.quad_step_backward(exp_x, upper_exp_y, upper_adj_v, t, quad_method=QuadMethod.Simpson)

        if lower_adj_x is not None:
            lower_exp_x = self.exp_transfer(lower_adj_x)
            lower_v = self.quad_step_backward(lower_exp_x[:2], y, v, t, quad_method=QuadMethod.Simpson)
            if lower_adj_y is not None:
                lower_v += self.quad_step_backward(lower_exp_x[:2], lower_exp_y, lower_adj_v, t,
                                                   quad_method=QuadMethod.Simpson)
            if upper_adj_y is not None:
                lower_v += self.quad_step_backward(lower_exp_x[:2], upper_exp_y, upper_adj_v, t,
                                                   quad_method=QuadMethod.Simpson)
        else:
            lower_v = None

        if upper_adj_x is not None:
            upper_exp_x = self.exp_transfer(upper_adj_x)
            upper_v = self.quad_step_backward(upper_exp_x[-2:], y, v, t, quad_method=QuadMethod.Simpson)
            if lower_adj_y is not None:
                upper_v += self.quad_step_backward(upper_exp_x[-2:], lower_exp_y, lower_adj_v, t,
                                                   quad_method=QuadMethod.Simpson)
            if upper_adj_y is not None:
                upper_v += self.quad_step_backward(upper_exp_x[-2:], upper_exp_y, upper_adj_v, t,
                                                   quad_method=QuadMethod.Simpson)
        else:
            upper_v = None

        return v_vec, lower_v, upper_v



#!/user/bin/env python
# -*- coding: utf-8 -*-
"""
Copyright (C) 2024 Galaxy Technologies
Licensed under the Apache License, Version 2.0
"""
import numpy as np

from ..utilities.enums import VolType
from ..utilities.patterns import Observable


class LocalVolSurface(Observable):
    vol_type = VolType.LV

    def __init__(self, expirations: np.ndarray, strikes: np.ndarray,
                 impliedvol: np.ndarray = None, volval: np.ndarray = None):
        """局部波动率曲面
        ivol: BS隐含波动率曲面矩阵
        expirations；时间格点array
        strikes：价格格点array
        lvol: 局部波动率曲面矩阵，行对应时间t，列对应价格s
        """
        super().__init__()
        self.impliedvol = impliedvol
        self.expirations = expirations  # 一维array, 时间格点, 升序排列
        self.strikes = strikes  # 二维矩阵，每行是时间，每列是执行价，执行价必须是升序排列的
        self.volval = volval  # np.array, 存储局部波动率矩阵

    def __call__(self, t, s):
        """返回指定时间t和制定价格s对应的局部波动率。t是float，s是array"""
        # 在已排序的数组self.t_grids中查找插入t的index，以维持数组的排序
        t_idx = np.searchsorted(self.expirations[:-1], t)
        s_idx = np.argmin(np.abs(self.strikes - s.reshape(-1, 1)), axis=1)
        lv = self.volval[t_idx, s_idx]
        return lv

    def dupire_formula(self):
        """TODO: 用Dupire公式计算局部波动率曲面"""



#!/user/bin/env python
# -*- coding: utf-8 -*-
"""
Copyright (C) 2024 Galaxy Technologies
Licensed under the Apache License, Version 2.0
"""
from .analytic_vanilla_european_engine import *
from .analytic_vanilla_american_engine import *
from .analytic_portfolio_engine import *
from .analytic_cash_or_nothing_engine import *
from .analytic_asset_or_nothing_engine import *
from .analytic_barrier_engine import *
from .analytic_asian_engine import *
from .analytic_double_digital_engine import *
from .analytic_double_barrier_engine import *
from .analytic_double_shark_engine import *
from .analytic_airbag_engine import *
from .cashflow_engine import *
from .analytic_heston_vanilla_engine import *

__all__ = ['AnalyticVanillaEuEngine', 'AnalyticBarrierEngine', 'AnalyticCashOrNothingEngine',
           'AnalyticDoubleDigitalEngine', 'AnalyticVanillaAmEngine', 'AnalyticAsianEngine', 'AnalyticPortfolioEngine',
           'AnalyticDoubleBarrierEngine', 'AnalyticDoubleSharkEngine', 'AnalyticAirbagEngine',
           'AnalyticAssetOrNothingEngine', 'CashFlowEngine', 'AnalyticHestonVanillaEngine', ]



#!/user/bin/env python
# -*- coding: utf-8 -*-
"""
Copyright (C) 2024 Galaxy Technologies
Licensed under the Apache License, Version 2.0
"""
import numpy as np
from pricelib.common.utilities.enums import AverageMethod, AsianAveSubstitution
from pricelib.common.pricing_engine_base import McEngine
from pricelib.common.time import global_evaluation_date


class MCAsianEngine(McEngine):
    """亚式期权 Monte Carlo 模拟定价引擎"""

    def calc_present_value(self, prod, t=None, spot=None):
        """计算现值
        Args:
            prod: Product产品对象
            t: datetime.date，估值日; 如果是None，则使用全局估值日globalEvaluationDate
            spot: float，估值日标的价格，如果是None，则使用随机过程的当前价格
        Returns: float，现值
        """
        calculate_date = global_evaluation_date() if t is None else t
        _maturity = (prod.end_date - calculate_date).days / prod.annual_days.value
        _maturity_business_days = prod.trade_calendar.business_days_between(calculate_date, prod.end_date)
        obs_start = prod.trade_calendar.business_days_between(calculate_date, prod.obs_start)
        obs_end = prod.trade_calendar.business_days_between(calculate_date, prod.obs_end)

        if spot is None:
            spot = self.process.spot()
        else:
            self.reset_paths_flag()  # 重置路径标志位，重新生成路径

        paths = self.path_generator(n_step=_maturity_business_days, spot=spot,
                                    t_step_per_year=prod.t_step_per_year).copy()
        r = self.process.interest(_maturity)
        obs_steps = np.arange(obs_start, obs_end, 1, dtype=int)
        if prod.substitute == AsianAveSubstitution.Underlying:
            if prod.enhanced:
                paths[obs_steps, :] = paths[obs_steps, :] + prod.callput.value * np.maximum(
                    prod.callput.value * (prod.limited_price - paths[obs_steps, :]), 0)
            if prod.ave_method == AverageMethod.Arithmetic:
                ave_s = np.mean(paths[obs_steps, :], axis=0)
            elif prod.ave_method == AverageMethod.Geometric:
                ave_s = np.exp(np.mean(np.log(paths[obs_steps, :]), axis=0))
            else:
                raise ValueError("无效的平均价计算方法，只支持几何平均/算术平均")
            price = np.mean(np.maximum(prod.callput.value * (ave_s - prod.strike), 0)) * np.exp(-r * _maturity)
            return price
        if prod.substitute == AsianAveSubstitution.Strike:
            if prod.ave_method == AverageMethod.Arithmetic:
                ave_s = np.mean(paths[obs_steps, :], axis=0)
            elif prod.ave_method == AverageMethod.Geometric:
                ave_s = np.exp(np.mean(np.log(paths[obs_steps, :]), axis=0))
            else:
                raise ValueError("无效的平均价计算方法，只支持几何平均/算术平均")
            price = np.mean(np.maximum(prod.callput.value * (paths[-1] - ave_s), 0)) * np.exp(-r * _maturity)
            return price
        raise ValueError("无效的substitute类型，只支持替代标的资产到期价格/替代行权价")



#!/user/bin/env python
# -*- coding: utf-8 -*-
"""
Copyright (C) 2024 Galaxy Technologies
Licensed under the Apache License, Version 2.0
"""
from contextlib import suppress
import numpy as np
from scipy.interpolate import interp1d
from pricelib.common.utilities.enums import EngineType, CallPut, InOut, PaymentType, StatusType
from pricelib.common.utilities.utility import descending_pairs_for_barrier
from pricelib.common.pricing_engine_base import FdmEngine, FdmGridwithBound, FdmGrid
from pricelib.common.time import global_evaluation_date


class FdmBarrierEngine(FdmEngine):
    """障碍期权PDE有限差分法定价引擎
    只支持美式观察(整个有效期观察是否触碰)；支持连续观察、均匀离散观察(默认为每日观察)；
    敲入现金返还为到期支付；敲出现金返还支持 立即支付/到期支付"""
    engine_type = EngineType.PdeEngine

    def __init__(self, stoch_process=None, s_step=800, n_smax=2, fdm_theta=1, *,
                 s=None, r=None, q=None, vol=None):
        """初始化有限差分法定价引擎
        Args:
            stoch_process: StochProcessBase随机过程
            s_step: int，价格步数
            n_smax: int，价格网格上界，设为n_smax倍初始价格
            fdm_theta: float，时间方向有限差分theta，0：explicit, 1: implicit, 0.5: Crank-Nicolson
        在未设置stoch_process时，(stoch_process=None)，会默认创建BSMprocess，需要输入以下变量进行初始化
            s: float，标的价格
            r: float，无风险利率
            q: float，分红/融券率
            vol: float，波动率
        """
        super().__init__(stoch_process, s_step=s_step, n_smax=n_smax, fdm_theta=fdm_theta, s=s, r=r, q=q, vol=vol)
        # 计算过程的中间变量
        self.prod = None  # Product产品对象
        self.bound = None  # 障碍价格区间
        self.rebate = None  # 补偿回报
        self.window = None  # 观察时间窗口
        self.fd_not_in = None  # 未敲入的有限差分网格FdmGrid对象
        self.fd_knockin = None  # 已敲入的有限差分网格FdmGrid对象
        self.lower = None  # 小于低障碍价的标的价格bool索引
        self.upper = None  # 大于高障碍价的标的价格bool索引
        self.outer = None  # 小于低障碍价或大于高障碍价，即在两个障碍价格之外的bool索引
        self.inner = None  # 大于低障碍价且小于高障碍价，即在两个障碍价格之间的bool索引
        self.smax = None  # 价格网格上界
        self.fn_callput = None  # 无障碍价格的边界条件
        self.fn_bound = None  # 障碍价格的边界条件

    def _init_boundary_condition(self, smax, maturity):
        """生成无障碍价格的边界条件
        在s=0处，call近似为0; 在smax处，put近似为0。根据看涨看跌平价 call-put = S*exp(-qT) - K*exp(-rT),
        可得对应的put(0) = K * exp(-rT), call(smax) = smax * exp(-qT) - K * exp(-rT)
        Args:
            smax: float, 有限差分网格上边界的价格最大值
            maturity: float, 到期时间
        Returns:
            self.fn_callput: List[Callable], 0~smax网格的边界条件，callput = self.fn_callput[0](s) or self.fn_callput[1](s)
            self.fn_bound: List[Callable], 障碍价格网格的边界条件，bound = self.fn_bound[0](s) or self.fn_bound[1](s)
        """
        if self.prod.callput == CallPut.Call:
            self.fn_callput = [lambda u: 0, lambda u: (smax * self.process.div.disc_factor(maturity, u) -
                                                       self.prod.strike * self.process.interest.disc_factor(maturity,
                                                                                                            u)) * self.prod.parti]
        else:
            self.fn_callput = [lambda u: self.prod.strike * self.process.interest.disc_factor(maturity, u)
                                         * self.prod.parti, lambda u: 0]

        # 生成障碍价格的Dirichlet边界条件： 第u天触碰障碍产生的补偿，折现到第u天的价值。
        # 返回一个包含两个lambda函数的列表，列表索引分别为[0, 1]，每个lambda函数的形参r=self.rebate中的r补偿值
        # 需要注意的是，敲出期权rebate[0]是下边界补偿，敲出期权rebate[1]是上边界补偿；敲入期权rebate[0]是敲入补偿，敲入期权rebate[1]没有用处。
        self.fn_bound = []
        for i, r in enumerate(self.rebate):
            if self.bound[i] in [-np.inf, np.inf]:  # 障碍价无穷，即障碍价没有设置
                if self.prod.inout == InOut.Out:  # 敲出期权，未设置障碍时，边界条件是香草期权的payoff
                    self.fn_bound.append(self.fn_callput[i])
                else:  # 敲入期权，未设置障碍时，边界条件是未敲入补偿rebate
                    self.fn_bound.append(lambda u: self.rebate[0] * self.process.interest.disc_factor(maturity, u))
            else:  # 这里只设置敲出期权的边界条件，因为敲入期权的边界条件是由已敲入的期权价值覆盖的
                if self.prod.inout == InOut.Out:
                    if self.prod.payment_type == PaymentType.Hit:  # 触碰障碍, 立即支付现金补偿
                        self.fn_bound.append(lambda u: r)
                    else:  # 'PaymentMethod.Expire' 触碰障碍后，到期时支付现金补偿
                        self.fn_bound.append(lambda u: r * self.process.interest.disc_factor(maturity, u))
                else:
                    self.fn_bound.append(self.fn_callput[i])

    def _init_terminal_condition(self, s_vec):
        """终止条件
        连续观察（网格上下界是障碍价格）：初始化终止时已敲入/未敲出的期权价值
        离散观察（网格上下界是0和smax）：初始化终止时的期权价值
        """
        yv = np.maximum(self.prod.callput.value * (s_vec - self.prod.strike), 0) * self.prod.parti
        if self.prod.discrete_obs_interval is None:  # 连续观察
            return yv  # 已敲入/未敲出 都是香草
        else:  # 离散观察
            if self.prod.inout == InOut.In:  # 敲入障碍 # 已敲入
                return yv
            else:  # 敲出障碍  # 未敲出香草，已敲出rebate
                yv[self.lower] = self.rebate[0]
                yv[self.upper] = self.rebate[1]
                return yv

    def _init_no_touch_terminal_condition(self, s_vec, p=0, maturity=0):
        """初始化障碍观察区间终点未敲入的期权价值
            连续观察（网格上下界是障碍价格）：此时未敲入期权价值全部是补偿rebate
            离散观察（网格上下界是0和smax）：此时将未敲入部分是rebate, 敲入部分是香草
        """
        if self.prod.discrete_obs_interval is None:  # 连续观察
            yv = s_vec * 0 + self.rebate[0] * self.process.interest.disc_factor(maturity, p)
        else:  # 离散观察
            yv = np.maximum(self.prod.callput.value * (s_vec - self.prod.strike), 0) * self.prod.parti
            yv[self.inner] = self.rebate[0]
        return yv

    # pylint: disable=too-many-locals, too-many-branches, too-many-locals
    def calc_present_value(self, prod, t=None, spot=None, bound=(None, None), rebate=(0, 0),
                           window=(None, None)):
        """计算现值，PDE有限差分定价，时间逆向迭代
        Args:
            prod: Product产品对象
            t: datetime.date，估值日; 如果是None，则使用全局估值日globalEvaluationDate
            spot: float，估值日标的价格，如果是None，则使用随机过程的当前价格
            bound: (lower_barrier, upper_barrier), 低障碍价/高障碍价，障碍价格绝对数值。如果是默认值None，即为无穷大或无穷小inf。
            rebate: (lower_rebate, upper_rebate), 对敲出期权，元组第一位是低障碍价下方的补偿，第二位是高障碍价上方的补偿，
                                                  对敲入期权，元组第一位是敲入补偿，第二位没有用处。
            window: Tuple/List[float]，障碍观察时间窗口，(start, end)，距离估值日的年化期限
                    如果是默认值None，说明观察区间是整个存续期: 估值日-终止日
        Returns: float，现值
        """
        self.prod = prod
        self.bound = np.array([(2 * i - 1) * np.inf if b is None else b for i, b in enumerate(bound)])  # 障碍价格区间
        self.rebate = np.array(rebate)  # 补偿回报
        self.window = np.array(window)  # 观察时间窗口

        calculate_date = global_evaluation_date() if t is None else t  # 估值日
        # 剩余到期时间（交易日天数）
        maturity = prod.trade_calendar.business_days_between(calculate_date, prod.end_date) / prod.t_step_per_year
        if spot is None:
            spot = self.process.spot()
        self.smax = self.n_smax * spot  # 价格网格上界, n_smax倍初始价格

        window_start = 0 if self.window[0] is None else max(self.window[0], 0)  # 观察时间窗口开始start时间
        window_end = maturity if self.window[1] is None else max(self.window[1], 0)  # 观察时间窗口截止end时间

        # 返回PDE系数组件abc的函数
        fn_pde_coef = self.process.get_fn_pde_coef(maturity, spot)

        # 如果障碍是连续观察，网格上下界应该是障碍价格，如果障碍是离散观察，网格上下界应该是0~4倍初始价格
        if prod.discrete_obs_interval is None:  # 连续观察
            bound_rate = (min(self.smax, self.bound[1]) - max(0, self.bound[0])) / (self.smax - 0)  # 障碍价格占价格网格上下界的比例
            fd_full = FdmGridwithBound(smax=self.smax, t_step_per_year=prod.t_step_per_year, s_step=self.s_step,
                                       fn_pde_coef=fn_pde_coef, fdm_theta=self.fdm_theta)  # 无障碍价格，[0, smax]
            fd_bound = FdmGridwithBound(smax=self.smax, t_step_per_year=prod.t_step_per_year,
                                        s_step=round(self.s_step * bound_rate),
                                        fn_pde_coef=fn_pde_coef, bound=self.bound,
                                        fdm_theta=self.fdm_theta)  # 有障碍价格, Dirichlet边界条件
            self._init_boundary_condition(self.smax, maturity)
            fd_full.set_boundary_condition(self.fn_callput)

            self.lower = fd_full.s_vec <= self.bound[0]  # 小于低障碍价的标的价格bool索引
            self.upper = fd_full.s_vec >= self.bound[1]  # 大于高障碍价的标的价格bool索引
            self.outer = self.lower | self.upper  # 小于低障碍价或大于高障碍价，即在两个障碍价格之外的bool索引

            time_pairs = descending_pairs_for_barrier(0, window_start, window_end,
                                                      maturity)  # 将一系列时间从大到小排序，去重，包装成相邻的降序对
            fdm = fd_full  # fdm是当前时间点的fdm，初始值是fd_full，后面会根据时间点的变化，切换成fd_full或fd_bound
            # 从到期日从后向前倒推，此时无论敲入还是敲出，未敲入和已敲出的payoff已经固定是rebate无需计算，只有已敲入或未敲出的期权还需要计算价值。
            # df_diri代表未敲入或已敲出。到期日的fd_full代表已敲入或未敲出。

            if prod.inout == InOut.In:  # 敲入障碍。
                # 初始化到期日已敲入payoff价值yv=max(callput * (F-K), 0)
                yv = self._init_terminal_condition(fdm.s_vec)
                # 从到期日开始向前迭代，fd_full是已敲入。fd_bound是未敲入。
                # 由于障碍观察区间之后，未敲入的期权价值固定是rebate，所以不需要计算fd_bound这一部分，只需要计算fd_full。
                for p, q in time_pairs:  # p是起点t{j}, q是终点t{j-1}, p>q由后向前倒推
                    if p == window_end:  # 进入障碍观察区间，开始计算未敲入的期权价值fd_bound，此时fd_full是已敲入的。
                        yb = yv  # 记录进入观察区间之前的yv，称为yb(boundaries)，yb是已敲入的期权价值
                        # 初始化p时刻(观察区间终点)仍然未敲入的fd_bound的期权价值向量yv，yv全部是未敲入的补偿rebate
                        yv = self._init_no_touch_terminal_condition(fd_bound.s_vec, p, maturity)
                        # 每步更新的fdm切换成fd_bound，因为未敲入的fd_bound的边界条件需要被已敲入的fd_full覆盖，所以要先算fd_full，再算fd_bound
                        fdm = fd_bound
                    if window_end >= p > window_start:  # 处于障碍观察期间内。注意p == w_end也算在内。
                        # 在观察区间内。此时fd_full和fd_bound都在更新: fd_full迭代yb，fd_bound迭代yv。将未敲入的fd_bound的边界条件用已敲入的fd_full覆盖。
                        yb, bc_fn = fd_full.evolve_with_interval(p, q, yb, self.bound)
                        for i, b in enumerate(self.bound):  # 如果向上/向下敲入未设置边界，则边界条件应该是未敲入状态
                            if b in [-np.inf, np.inf]:
                                bc_fn[i] = self.fn_bound[i]
                        fd_bound.set_boundary_condition(bc_fn)
                    if p == window_start:
                        # 退出观察区间，此时在障碍价格之外部分的yv覆盖为已敲入或已敲出的yb，每步只更新fd_full。
                        # 这里只更新fd_full的原因，是因为在观察区间起点之前，不可能敲入，所有的价格路径都是未敲入的状态；最早可能敲入的点在观察区间起点。
                        # 而在敲入观察区间第一天，在outer外的部分payoff会变成rebate，因此将障碍价格之外部分的yv覆盖为已敲入或已敲出的yb。
                        yv = fd_bound.functionize(yv)(fd_full.s_vec)
                        yv[self.outer] = yb[self.outer]
                        fdm = fd_full
                    yv = fdm.evolve_with_interval(p, q, yv)  # 每组(p, q)更新。如果在障碍观察区间内，更新fd_bound；在观察区间外，更新fd_full
            else:  # 敲出障碍
                # 初始化到期日未敲出payoff价值yv=max(callput * (F-K), 0)
                yv = self._init_terminal_condition(fdm.s_vec)
                fd_bound.set_boundary_condition(self.fn_bound)
                for p, q in time_pairs:  # 倒序时间对，p>=q
                    if p == window_end:
                        # 进入观察区间，后面的fd_full都是未敲出的，因为已敲出的期权已经失效。fd_bound也是未敲出，只是比df_full加上了敲出障碍的边界条件。
                        yv = fd_full.functionize(yv)(fd_bound.s_vec)
                        fdm = fd_bound
                    if p == window_start:
                        # 退出观察区间，前面的fd_full都是不可能敲出的。需要将yv的lower和upper部分替换为rebate。
                        # 因为敲出观察区间第一天，在outer的部分会立即变成敲出，payoff会变成rebate
                        yv = fd_bound.functionize(yv)(fd_full.s_vec)
                        yv[self.lower] = self.fn_bound[0](p)
                        yv[self.upper] = self.fn_bound[1](p)
                        fdm = fd_full
                    yv = fdm.evolve_with_interval(p, q, yv)  # 每组(p, q)更新。如果在障碍观察区间内，更新fd_bound；在观察区间外，更新fd_full

            return np.float64(fdm.functionize(yv)(spot))
        else:  # 离散观察
            t_step = round(prod.t_step_per_year * maturity)  # 时间步数
            dt = 1 / prod.t_step_per_year
            fn_pde_coef = self.process.get_fn_pde_coef(maturity, spot)
            self.fd_not_in = FdmGrid(smax=self.smax, t_step_per_year=prod.t_step_per_year, s_step=self.s_step,
                                     fn_pde_coef=fn_pde_coef, fdm_theta=self.fdm_theta,
                                     maturity=maturity)
            self._init_boundary_condition(smax=self.smax, maturity=maturity)
            self.fd_not_in.v_grid[0, :] = self.fn_bound[0](self.fd_not_in.tv)
            self.fd_not_in.v_grid[-1, :] = self.fn_bound[1](self.fd_not_in.tv)

            self.lower = self.fd_not_in.s_vec <= self.bound[0]  # 小于低障碍价的标的价格bool索引
            self.upper = self.fd_not_in.s_vec >= self.bound[1]  # 大于高障碍价的标的价格bool索引
            self.outer = self.lower | self.upper  # 小于低障碍价或大于高障碍价，即在两个障碍价格之外的bool索引
            self.inner = ~self.lower & ~self.upper

            obs_points = np.flip(np.round(prod.t_step_per_year *
                                          np.arange(maturity, 0, -prod.discrete_obs_interval)).astype(int))
            obs_points = np.concatenate((np.array([0]), obs_points))

            if prod.inout == InOut.In:  # 敲入障碍
                self.fd_knockin = FdmGrid(smax=self.smax, t_step_per_year=prod.t_step_per_year, s_step=self.s_step,
                                          fn_pde_coef=fn_pde_coef, fdm_theta=self.fdm_theta, maturity=maturity)
                self.fd_knockin.v_grid[-1, :] = self.fn_callput[1](self.fd_knockin.tv)
                self.fd_knockin.v_grid[0, :] = self.fn_callput[0](self.fd_knockin.tv)
                # 初始化到期日payoff价值
                yv = self._init_no_touch_terminal_condition(self.fd_not_in.s_vec)
                yv_in = self._init_terminal_condition(self.fd_knockin.s_vec)

                self.fd_not_in.v_grid[1:-1, -1] = yv
                self.fd_knockin.v_grid[1:-1, -1] = yv_in

                for j in range(t_step - 1, -1, -1):
                    yv = self.fd_not_in.evolve(j, yv, dt)  # 每组(p, q)更新。如果在障碍观察区间内，更新fd_bound；在观察区间外，更新fd_full
                    yv_in = self.fd_knockin.evolve(j, yv_in, dt)
                    if j in obs_points:
                        yv[self.outer] = yv_in[self.outer]
                    self.fd_not_in.v_grid[1:-1, j] = yv
                    self.fd_knockin.v_grid[1:-1, j] = yv_in
            else:  # 敲出障碍
                # 初始化到期日payoff价值
                yv = self._init_terminal_condition(self.fd_not_in.s_vec)
                self.fd_not_in.v_grid[1:-1, -1] = yv
                for j in range(t_step - 1, -1, -1):
                    yv = self.fd_not_in.evolve(j, yv, dt)  # 每组(p, q)更新。如果在障碍观察区间内，更新fd_bound；在观察区间外，更新fd_full
                    if j in obs_points:
                        if prod.payment_type == PaymentType.Expire:  # 到期支付现金补偿
                            yv[self.lower] = self.fn_bound[0](j / t_step * maturity)
                            yv[self.upper] = self.fn_bound[1](j / t_step * maturity)
                        else:  # 立即支付现金补偿
                            yv[self.lower] = self.fn_bound[0](maturity)
                            yv[self.upper] = self.fn_bound[1](maturity)
                    self.fd_not_in.v_grid[1:-1, j] = yv
            return np.float64(self.fd_not_in.functionize(yv, kind="cubic")(spot))

    def delta(self, prod, t=0, spot=None, step=None, status: StatusType = None):
        """求t时刻，价格spot处的delta值
        Args:
            t: int，期权价值矩阵的列(时间)的索引
            spot: float，价格的绝对值
            step: float，微扰步长，默认为价格spot的1%
            status: StatusType，期权是否已敲入
        Returns: delta = ∂V/∂S
        """
        spot = self.process.spot() if spot is None else spot
        step = spot * 0.01 if step is None else step
        status = self.prod.status if status is None else status
        fdm_grid = self.fd_knockin if (status != StatusType.NoTouch) else self.fd_not_in
        f = fdm_grid.functionize(fdm_grid.v_grid[1:-1, t], kind='cubic')
        delta = (f(min(spot + step, self.smax)) - f(min(spot - step, self.smax))) / (2 * step)
        return delta

    def gamma(self, prod, t=0, spot=None, step=None, status: StatusType = None):
        """求t时刻，价格spot处的gamma值
        Args:
            t: int，期权价值矩阵的列(时间)的索引
            spot: float，价格的绝对值
            step: float，微扰步长，默认为价格spot的1%
            status: StatusType，期权是否已敲入
        Returns: gamma = ∂2V/∂S2
        """
        spot = self.process.spot() if spot is None else spot
        step = spot * 0.01 if step is None else step
        status = self.prod.status if status is None else status
        fdm_grid = self.fd_knockin if (status != StatusType.NoTouch) else self.fd_not_in
        f = interp1d(fdm_grid.s_vec, fdm_grid.v_grid[1:-1, t], kind='cubic')
        gamma = (f(spot + step) - 2 * f(spot) + f(spot - step)) / (step ** 2)
        return gamma

    def theta(self, prod, t=0, spot=None, step=1, status: StatusType = None):
        """计算一天的theta值
        Args:
            t: int，期权价值矩阵的列(时间)的索引
            spot: float，价格的绝对值
            step: int，时间步长，默认为1天
            status: StatusType，期权是否已敲入
        Returns: theta = ∂V/∂t
        """
        spot = self.process.spot() if spot is None else spot
        status = self.prod.status if status is None else status
        fdm_grid = self.fd_knockin if (status != StatusType.NoTouch) else self.fd_not_in
        f = fdm_grid.functionize(fdm_grid.v_grid[1:-1, t], kind='cubic')
        g = fdm_grid.functionize(fdm_grid.v_grid[1:-1, t + step], kind='cubic')
        theta = (g(spot) - f(spot)) / (step * spot)  # ( step * self.dt)
        return theta

    def vega(self, prod, t=0, spot=None, step=0.01, status: StatusType = None):
        """计算波动率上升1%的vega todo:目前只支持常数波动率
        Args:
            t: int，期权价值矩阵的列(时间)的索引
            spot: float，价格的绝对值
            step: float，微扰步长，默认为1%
            status: StatusType，期权是否已敲入
        Returns: vega = ∂V/∂σ
        """
        spot = self.process.spot() if spot is None else spot
        status = self.prod.status if status is None else status
        last_vol = self.process.vol.volval
        last_in_grid = self.fd_knockin.v_grid.copy()
        last_not_grid = self.fd_not_in.v_grid.copy()
        fdm_grid0 = self.fd_knockin if (status != StatusType.NoTouch) else self.fd_not_in
        f = fdm_grid0.functionize(fdm_grid0.v_grid[1:-1, t], kind='cubic')
        self.process.vol.volval = last_vol + step
        self.calc_present_value(prod=prod)
        fdm_grid1 = self.fd_knockin if (status == StatusType.DownTouch) else self.fd_not_in
        g = fdm_grid1.functionize(fdm_grid1.v_grid[1:-1, t], kind='cubic')
        vega = (g(spot) - f(spot)) / spot
        self.process.vol.volval = last_vol
        self.fd_knockin.v_grid = last_in_grid
        self.fd_not_in.v_grid = last_not_grid
        return vega

    def rho(self, prod, t=0, spot=None, step=0.01, status: StatusType = None):
        """计算无风险利率上升1%的rho todo:目前只支持常数无风险利率
        Args:
            t: int，期权价值矩阵的列(时间)的索引
            spot: float，价格的绝对值
            step: float，微扰步长，默认为1%
            status: StatusType，期权是否已敲入
        Returns: rho = ∂V/∂r
        """
        spot = self.process.spot() if spot is None else spot
        status = self.prod.status if status is None else status
        last_r = self.process.interest.data
        last_in_grid = self.fd_knockin.v_grid.copy()
        last_not_grid = self.fd_not_in.v_grid.copy()
        fdm_grid0 = self.fd_knockin if (status != StatusType.NoTouch) else self.fd_not_in
        f = fdm_grid0.functionize(fdm_grid0.v_grid[1:-1, t], kind='cubic')
        self.process.interest.data = last_r + step
        self.calc_present_value(prod=prod)
        fdm_grid1 = self.fd_knockin if (status == StatusType.DownTouch) else self.fd_not_in
        g = fdm_grid1.functionize(fdm_grid1.v_grid[1:-1, t], kind='cubic')
        rho = (g(spot) - f(spot)) / spot
        self.process.interest.data = last_r
        self.fd_knockin.v_grid = last_in_grid
        self.fd_not_in.v_grid = last_not_grid
        return rho

    def pv_and_greeks(self, prod, t=0, spot=None, status: StatusType = None):
        """当一次性计算pv和5个greeks时，可以调用此函数，减少计算量
        Args:
            t: int，期权价值矩阵的列(时间)的索引
            spot: float，价格的绝对值
            status: StatusType，期权是否已敲入
        Returns:
            Dict[str:float]: {'pv': pv, 'delta': delta, 'gamma': gamma, 'theta': theta, 'vega': vega, 'rho': rho}
        """
        spot = self.process.spot() if spot is None else spot
        status = self.prod.status if status is None else status
        s_step = spot * 0.01
        fdm_grid = self.fd_knockin if (status != StatusType.NoTouch) else self.fd_not_in
        f = fdm_grid.functionize(fdm_grid.v_grid[1:-1, t], kind='cubic')
        pv = np.float64(f(spot))
        delta = (f(min(spot + s_step, self.smax)) - f(min(spot - s_step, self.smax))) / (2 * s_step)
        gamma = (f(spot + s_step) - 2 * f(spot) + f(spot - s_step)) / (s_step ** 2)

        g = fdm_grid.functionize(fdm_grid.v_grid[1:-1, t + 1], kind='cubic')
        theta = (g(spot) - f(spot)) / (1 * spot)
        with suppress(AttributeError):
            last_in_grid = self.fd_knockin.v_grid.copy()
        last_not_grid = self.fd_not_in.v_grid.copy()

        last_vol = self.process.vol.volval
        self.process.vol.volval = last_vol + 0.01
        self.calc_present_value(prod=prod)
        fdm_grid1 = self.fd_knockin if (status != StatusType.NoTouch) else self.fd_not_in
        g1 = fdm_grid1.functionize(fdm_grid1.v_grid[1:-1, t], kind='cubic')
        vega = (g1(spot) - f(spot)) / spot
        self.process.vol.volval = last_vol

        last_r = self.process.interest.data
        self.process.interest.data = last_r + 0.01
        self.calc_present_value(prod=prod)
        fdm_grid2 = self.fd_knockin if (status != StatusType.NoTouch) else self.fd_not_in
        g2 = fdm_grid2.functionize(fdm_grid2.v_grid[1:-1, t], kind='cubic')
        rho = (g2(spot) - f(spot)) / spot
        self.process.interest.data = last_r
        with suppress(AttributeError, UnboundLocalError):
            self.fd_knockin.v_grid = last_in_grid
        self.fd_not_in.v_grid = last_not_grid

        return {'pv': pv, 'delta': delta, 'gamma': gamma, 'vega': vega, 'theta': theta, 'rho': rho}

    def value_matrix(self, status: StatusType = None):
        """根据是否敲入状态，返回S-t矩阵
        Args:
            status: StatusType，期权是否已敲入
        Returns:
            v_grid: np.ndarray，S-t矩阵
            s_vec: np.ndarray，S轴格点
        """
        status = self.prod.status if status is None else status
        fdm_grid = self.fd_knockin if (status == StatusType.DownTouch) else self.fd_not_in
        return fdm_grid.v_grid, fdm_grid.s_vec

    def delta_matrix(self, step=0, status: StatusType = None):
        """计算S-t矩阵中，每个t时刻所有s格点的delta
        Args:
            step: float, 标的价步长，默认为0，即选择PDE有限差分网格的价格步长
            status: StatusType，期权是否已敲入
        Returns:
            delta_s_t: np.ndarray，Delta-t矩阵
            s_vec: np.ndarray，S轴格点
        """
        status = self.prod.status if status is None else status
        fdm_grid = self.fd_knockin if (status == StatusType.DownTouch) else self.fd_not_in
        if step == 0:
            # 默认PDE模拟时的步长 step = fdm_grid.ds
            delta_s_t = (fdm_grid.v_grid[2:] - fdm_grid.v_grid[:-2]) / (2 * fdm_grid.ds)
            return delta_s_t, fdm_grid.s_vec  # 中心差分，价格向量掐头0去尾smax，正好是fdm_grid.s_vec
        else:
            # 重新定义步长，在模拟矩阵的基础上插值
            n_step = np.round(self.smax / step).astype('int')
            spots = np.arange(0, n_step, 1) * step
            # 下方n_step和spots用于避免计算远离敲入水平的场景，降低运算量
            # n_step = 2 * np.round((self.spot - self.barrier_in + 1) / step)
            # spots = np.arange(0, n_step, 1) * step + self.barrier_in - 1
            delta_s_t = np.zeros((n_step - 1, fdm_grid.v_grid.shape[1]))
            # 分别在fdm_grid.s_vec价格向量首位添加0和smax
            x_vec = np.insert(fdm_grid.s_vec, 0, 0)
            x_vec = np.append(x_vec, self.smax)
            for j in range(fdm_grid.v_grid.shape[1]):
                f = interp1d(x_vec, fdm_grid.v_grid[:, j], kind='cubic')
                for i, spot in enumerate(spots[1:]):
                    delta_s_t[i, j] = (f(spot + step) - f(spot - step)) / (step * 2)
            return delta_s_t, spots

    def gamma_matrix(self, step=0, status: StatusType = None):
        """计算S-t矩阵中，每个t时刻所有s格点的gamma
        Args:
            step: float, 标的价步长，默认为0，即选择PDE有限差分网格的价格步长
            status: StatusType，期权是否已敲入
        Returns:
            gamma_s_t: np.ndarray，Gamma-t矩阵
            s_vec: np.ndarray，S轴格点
        """
        status = self.prod.status if status is None else status
        fdm_grid = self.fd_knockin if (status == StatusType.DownTouch) else self.fd_not_in
        if step == 0:
            # 默认PDE模拟时的步长 step = fdm_grid.ds
            gamma_s_t = np.diff(fdm_grid.v_grid, n=2, axis=0) / fdm_grid.ds ** 2
            return gamma_s_t, fdm_grid.s_vec
        else:
            # 重新定义步长，在模拟矩阵的基础上插值
            n_step = np.round(self.smax / step).astype('int')
            step = self.smax / n_step
            spots = np.arange(0, n_step, 1) * step

            gamma_s_t = np.zeros((n_step - 2, fdm_grid.v_grid.shape[1]))
            # 分别在fdm_grid.s_vec价格向量首位添加0和smax
            x_vec = np.insert(fdm_grid.s_vec, 0, 0)
            x_vec = np.append(x_vec, self.smax)
            for j in range(fdm_grid.v_grid.shape[1]):
                f = interp1d(x_vec, fdm_grid.v_grid[:, j], kind='cubic')
                for i, spot in enumerate(spots[1:-2]):
                    gamma_s_t[i, j] = (f(spot + step) - 2 * f(spot) + f(spot - step)) / (step ** 2)
            return gamma_s_t, spots

    def theta_matrix(self, step=0, status: StatusType = None):
        """计算S-t矩阵中，每个s价格所有t时刻的theta
        Args:
            step: int, 时间步长，默认为0，即选择PDE有限差分网格的时间步长
            status: StatusType，期权是否已敲入
        Returns:
            theta_s_t: np.ndarray，Theta-t矩阵
            s_vec: np.ndarray，S轴格点
        """
        status = self.prod.status if status is None else status
        fdm_grid = self.fd_knockin if (status == StatusType.DownTouch) else self.fd_not_in
        if step == 0:
            # 默认PDE模拟时的步长
            spot = self.process.spot()
            theta_s_t = np.diff(fdm_grid.v_grid, axis=1) / spot  # / self.dt
            x_vec = np.insert(fdm_grid.s_vec, 0, 0)
            x_vec = np.append(x_vec, self.smax)
            return theta_s_t, x_vec
        else:  # 重新定义步长，在模拟矩阵的基础上插值
            raise NotImplementedError("todo: 时间步长与pde有限差分网格不同的theta.")



#!/user/bin/env python
# -*- coding: utf-8 -*-
"""
Copyright (C) 2024 Galaxy Technologies
Licensed under the Apache License, Version 2.0
"""
from .curve import *

__all__ = ['ConstantRate', 'RateTermStructure', 'DiscountCurve']



#!/user/bin/env python
# -*- coding: utf-8 -*-
"""
Copyright (C) 2024 Galaxy Technologies
Licensed under the Apache License, Version 2.0
"""
import datetime
import numpy as np
from pricelib.common.utilities.enums import StatusType
from pricelib.common.time import Schedule, CN_CALENDAR, AnnualDays
from pricelib.common.utilities.utility import time_this
from pricelib.pricing_engines.mc_engines import MCParisSnowballEngine
from .autocallable_base import AutocallableBase


class ParisSnowball(AutocallableBase):
    """巴黎雪球产品类
    与经典雪球相比，敲出观察频率变为每周观察，敲入观察频率变为每月观察，并且累计敲入两次才算敲入"""

    def __init__(self, s0, barrier_out, barrier_in, coupon_out, lock_term=3, obs_dates=None, pay_dates=None,
                 coupon_div=None, knock_in_obs_dates=None, knock_in_times=2, knockout_freq="w", parti_in=1,
                 margin_lvl=1, status=StatusType.NoTouch, engine=None, maturity=None, start_date=None, end_date=None,
                 trade_calendar=CN_CALENDAR, annual_days=AnnualDays.N365, t_step_per_year=243,
                 s=None, r=None, q=None, vol=None):
        """继承自动赎回基类AutocallableBase的参数，详见AutocallableBase的__init__方法
        Args: 以下为区别于经典雪球的参数
            knock_in_obs_dates: 敲入观察日期列表，List[datetime.date]，可缺省，缺省时会自动生成每月观察的敲入日期序列(已根据节假日调整)
            knock_in_times: 巴黎特性 - 累计敲入n次才算敲入
            knockout_freq: 敲出观察频率，"w"表示每周观察，"m"表示每月观察
        可选参数:
            若未提供引擎的情况下，提供了标的价格、无风险利率、分红/融券率、波动率，
            则默认使用巴黎雪球蒙特卡洛模拟定价引擎 FdmSnowBallEngine
            s: float，标的价格
            r: float，无风险利率
            q: float，分红/融券率
            vol: float，波动率
        """
        super().__init__(s0=s0, maturity=maturity, start_date=start_date, end_date=end_date, lock_term=lock_term,
                         trade_calendar=trade_calendar, status=status, annual_days=annual_days,
                         t_step_per_year=t_step_per_year, parti_in=parti_in, margin_lvl=margin_lvl)
        # 巴黎雪球参数
        if knock_in_obs_dates is None:  # 敲入观察日列表，每月观察
            self.knock_in_obs_dates = Schedule(trade_calendar=trade_calendar, start=self.start_date, end=self.end_date,
                                               freq='m', lock_term=1)
            self.end_date = self.knock_in_obs_dates.end  # 修正结束时间
        else:
            self.knock_in_obs_dates = Schedule(trade_calendar=trade_calendar, date_schedule=knock_in_obs_dates)
        assert knockout_freq in ["w",
                                 "m"], f'Error: 敲出观察频率仅支持每周或每月观察，knockout_freq当前为{knockout_freq}，请填"w"或"m"!'
        self.knockout_freq = knockout_freq
        if obs_dates is None:  # 敲出观察日列表，每周观察/每周观察
            self.obs_dates = Schedule(trade_calendar=trade_calendar, start=self.start_date, end=self.end_date,
                                      freq=self.knockout_freq, lock_term=lock_term)
        else:
            self.obs_dates = Schedule(trade_calendar=trade_calendar, date_schedule=obs_dates)
        if pay_dates is None:  # 付息日
            self.pay_dates = self.obs_dates
        else:
            self.pay_dates = Schedule(trade_calendar=trade_calendar, date_schedule=pay_dates)

        self.knock_in_times = knock_in_times  # 多少次敲入视为敲入，默认为2
        # 平敲参数
        len_obs_dates = len(self.obs_dates.date_schedule)
        self.barrier_out = np.ones(len_obs_dates) * barrier_out
        self.barrier_in = np.ones(len(self.knock_in_obs_dates.date_schedule)) * barrier_in
        self.coupon_out = np.ones(len_obs_dates) * coupon_out
        self.coupon_div = coupon_div if coupon_div is not None else self.coupon_out[-1]
        self.parti_out = 0
        self.strike_upper = s0
        self.strike_lower = 0
        if engine is not None:
            self.set_pricing_engine(engine)
        elif s is not None and r is not None and q is not None and vol is not None:
            default_engine = MCParisSnowballEngine(s=s, r=r, q=q, vol=vol)
            self.set_pricing_engine(default_engine)

    @time_this
    def price(self, t: datetime.date = None, spot=None):
        """计算期权价格
        Args:
            t: datetime.date，计算期权价格的日期
            spot: float，标的价格
        Returns: 期权现值
        """
        self.validate_parameters(t=t)
        price = self.engine.calc_present_value(prod=self, t=t, spot=spot)
        return price

    def __repr__(self):
        """返回期权的描述"""
        return "巴黎雪球"



#!/user/bin/env python
# -*- coding: utf-8 -*-
"""
Copyright (C) 2024 Galaxy Technologies
Licensed under the Apache License, Version 2.0
"""
from .patterns import Observer, Observable, SimpleQuote, HashableArray
from .utility import time_this, logging, ascending_pairs, descending_pairs
from .numerical import LinearFlat, CubicSplineFlat, FlatCubicSpline, TDMA_ldu_jit
from .enums import *


__all__ = ['Observer', 'Observable', 'SimpleQuote', 'HashableArray', 'time_this', 'logging',
           'ascending_pairs', 'descending_pairs',
           'LinearFlat', 'CubicSplineFlat', 'FlatCubicSpline', 'TDMA_ldu_jit',
           'CallPut', 'BuySell', 'RandsMethod', 'LdMethod', 'VolType', 'ProcessType', 'EngineType', 'QuadMethod',
           'BarrierType', 'UpDown', 'InOut', 'TouchType']



#!/user/bin/env python
# -*- coding: utf-8 -*-
"""
Copyright (C) 2024 Galaxy Technologies
Licensed under the Apache License, Version 2.0
"""
import math
import numpy as np
from pricelib.common.utilities.enums import CallPut, ExerciseType, PaymentType
from pricelib.common.time import global_evaluation_date
from pricelib.common.pricing_engine_base import BiTreeEngine


class BiTreeDigitalEngine(BiTreeEngine):
    """二元(数字)期权PDE二叉树定价引擎"""

    def calc_present_value(self, prod, t=None, spot=None):
        """计算现值
        Args:
            prod: Product产品对象
            t: datetime.date，估值日; 如果是None，则使用全局估值日globalEvaluationDate
            spot: float，估值日标的价格，如果是None，则使用随机过程的当前价格
        Returns: float，现值
        """
        calculate_date = global_evaluation_date() if t is None else t
        tau = prod.trade_calendar.business_days_between(calculate_date, prod.end_date) / prod.t_step_per_year
        if spot is None:
            spot = self.process.spot()

        self.r = self.process.interest(tau)
        n_step = round(tau * self.tree_branches)
        s_paths = self.path_generator(n_step, tau, spot).copy()
        v_grid = np.zeros(shape=s_paths.shape)

        # 欧式：
        if prod.exercise_type == ExerciseType.European:
            for j in range(n_step, -1, -1):
                if j == n_step:
                    if prod.callput == CallPut.Call:
                        v_grid[:, -1] = (s_paths[:, -1] >= prod.strike) * prod.rebate
                    elif prod.callput == CallPut.Put:
                        v_grid[:, -1] = (s_paths[:, -1] <= prod.strike) * prod.rebate
                else:
                    v_grid[:-1, j] = v_grid[:-1, j + 1] * self.p + v_grid[1:, j + 1] * (1 - self.p)
                    v_grid[:-1, j] *= np.exp(-self.r * self.dt)

        # 美式：
        if prod.exercise_type == ExerciseType.American:
            # 设置观察时点obs_points
            if prod.discrete_obs_interval is None:
                obs_points = np.arange(0, n_step + 1, 1)[1:]
            else:
                obs_points = np.flip(np.round(np.arange(tau, 0, -prod.discrete_obs_interval) * self.tree_branches))
                obs_points = np.concatenate((np.array([0]), obs_points))
            obs_points = np.round(obs_points).astype(int)

            for j in range(n_step, -1, -1):
                if j == n_step:
                    if prod.callput == CallPut.Call:
                        v_grid[:, -1] = (s_paths[:, -1] >= prod.strike) * prod.rebate
                    elif prod.callput == CallPut.Put:
                        v_grid[:, -1] = (s_paths[:, -1] <= prod.strike) * prod.rebate
                    else:
                        raise ValueError("Unknown CallPut type")
                else:
                    v_grid[:j + 1, j] = v_grid[:j + 1, j + 1] * self.p + v_grid[1:j + 2, j + 1] * (1 - self.p)
                    v_grid[:-1, j] *= np.exp(-self.r * self.dt)
                    if j in obs_points:
                        if prod.callput == CallPut.Call:
                            upper_s = np.where(self.s_paths[:j, j] >= prod.strike)
                            if prod.payment_type == PaymentType.Hit:
                                v_grid[upper_s, j] = prod.rebate
                            elif prod.payment_type == PaymentType.Expire:
                                time_remain = self.dt * (n_step - j)
                                v_grid[upper_s, j] = prod.rebate * math.exp(-self.r * time_remain)

                        elif prod.callput == CallPut.Put:
                            lower_s = np.where(self.s_paths[:j, j] <= prod.strike)
                            if prod.payment_type == PaymentType.Hit:
                                v_grid[lower_s, j] = prod.rebate
                            elif prod.payment_type == PaymentType.Expire:
                                time_remain = self.dt * (n_step - j)
                                v_grid[lower_s, j] = prod.rebate * math.exp(-self.r * time_remain)
        return v_grid[0, 0]



#!/user/bin/env python
# -*- coding: utf-8 -*-
"""
Copyright (C) 2024 Galaxy Technologies
Licensed under the Apache License, Version 2.0
"""
from .constantvol import *
from .localvol import *
from .stochvolatility import StochLocalVol, HestonSLV

__all__ = ['BlackConstVol', 'LocalVolSurface']



#!/user/bin/env python
# -*- coding: utf-8 -*-
"""
Copyright (C) 2024 Galaxy Technologies
Licensed under the Apache License, Version 2.0
"""
import numpy as np
from numba import njit
from ..utilities.enums import ProcessType
from ..utilities.patterns import Observable, Observer
from .stoch_process import StochProcessBase


@njit(parallel=True, fastmath=True, cache=False)
def evolve_jit(x0, dt, dw, var_kappa, var_theta, var_vol, drift):
    """Heston SDE演化函数，jit加速
    Heston方差非负修正 - 采用部分截断模式
    Args:
        x0: np.ndarray, [s0, v0]是上一时刻的标的价格和方差, 即s0 = x0[0], v0 = x0[1]
        dt: float, 年化时间增量
        dw: np.ndarray, [dw_s, dw_v]是s和v的布朗运动增量，这里只是标准正态分布随机数，函数内部会 * sqrt(dt)
        var_kappa: float，Heston参数，方差回归速度
        var_theta: float，Heston参数，方差均值
        var_vol: float，Heston参数，方差的波动率
        drift: float，s的漂移率
    Returns:
        [s1, v1], t0+dt时刻的标的价格和方差np.ndarray
    """
    v1 = x0[1] + var_kappa * (var_theta - x0[1]) * dt + var_vol * np.sqrt(
        np.maximum(x0[1], 0) * dt) * dw[1]
    s1 = x0[0] * (1 + drift * dt + dw[0] * np.sqrt(np.maximum(x0[1], 0) * dt))
    return [s1, v1]


class HestonProcess(StochProcessBase, Observer, Observable):
    """Heston SDE:
        dS/S = (r-q)dt + sqrt(v) dW
        dv = kappa(theta - v)dt + var_vol * sqrt(v) dZ
        E[dW dZ] = rho dt
    既是观察者又是被观察者，观察者模式的中介者。观察S、r、q、heston参数的变化，被定价引擎engine观察。
    """
    process_type = ProcessType.Heston

    def __init__(self, spot: Observable, interest: Observable, div: Observable, *,
                 v0: float, var_theta: float, var_kappa: float, var_vol: float, var_rho: float):
        """初始化HestonProcess实例
        Args:
            spot: 市场行情实例，标的价格
            interest: 市场行情实例，无风险利率
            div: 市场行情实例，分红融券率
            v0: float, 方差初始值
            var_theta: float, 方差均值
            var_kappa: float, 方差均值回归速率
            var_vol: float, 方差的波动率
            var_rho: float, 方差与标的资产布朗运动的相关系数
        """
        self.spot = spot  # 标的价格
        self.interest = interest  # 无风险利率
        self.div = div  # 分红融券率
        spot.add_observer(self)  # 将本实例加入spot观察者列表
        interest.add_observer(self)  # 将本实例加入interest观察者列表
        div.add_observer(self)  # 将本实例加入div观察者列表
        self._v0 = v0  # 方差初始值
        self._var_kappa = var_kappa  # 回归速率
        self._var_theta = var_theta  # 方差均值
        self._var_vol = var_vol  # 方差的波动率
        self._var_rho = var_rho  # 方差与标的资产布朗运动的相关系数
        super().__init__()  # 调用被观察者Observable的构造函数

    def remove_self(self):
        """删除对象自己，del自己之前，先将自己从被观察者的列表中移除"""
        self.spot.remove_observer(self)
        self.interest.remove_observer(self)
        self.div.remove_observer(self)
        del self

    def update(self, observable, *args, **kwargs):
        """被观察者的值发生变化时，自动调用update方法，通知观察者"""
        self.notify_observers()  # 自动向观察者发送通知, 参数已改变

    def __call__(self):
        return ProcessType.Heston

    # @property
    def drift(self, t):
        """Heston过程价格s的漂移项
        Args:
            t: float，距离起始日的年化时间
        Returns:
            drift: float = r - q，r是无风险利率，q是分红融券率
        """
        return self.interest(t) - self.div(t)

    def diffusion(self):
        """heston的扩散项是sqrt(v), 此处方差v是evolve方法的参数，从外界传入，todo: process暂时未存储s和v路径。"""

    def evolve(self, t, x, dt, dw):
        """演化函数，根据Heston SDE进行演化
        Args:
            t: float，距离起始日的年化时间
            x: np.ndarray, [s0, v0]是上一时刻的标的价格和方差, 即s0 = x0[0], v0 = x0[1]
            dt: float, 年化时间增量
            dw: np.ndarray, [dw_s, dw_v]是s和v的布朗运动增量，这里只是标准正态分布随机数，函数内部会 * sqrt(dt)
        Returns:
            [s1, v1], t0+dt时刻的标的价格和方差np.ndarray
        """
        return evolve_jit(np.array(x), dt, np.array(dw), self.var_kappa, self.var_theta, self.var_vol, self.drift(t))

    def get_fn_pde_coef(self):
        raise NotImplementedError("TODO: Heston PDE数值解法 - ADI 交替隐式迭代法")

    @property
    def v0(self):
        return self._v0

    @v0.setter
    def v0(self, new_value):
        """覆写给对象的属性赋值的方法，当属性值变化时，自动向观察者发送通知"""
        self._v0 = float(new_value)  # 属性变化
        self.notify_observers()  # 自动向观察者发送通知, 参数已改变

    @property
    def var_kappa(self):
        return self._var_kappa

    @var_kappa.setter
    def var_kappa(self, new_value):
        """覆写给对象的属性赋值的方法，当属性值变化时，自动向观察者发送通知"""
        self._var_kappa = float(new_value)  # 属性变化
        self.notify_observers()  # 自动向观察者发送通知, 参数已改变

    @property
    def var_theta(self):
        return self._var_theta

    @var_theta.setter
    def var_theta(self, new_value):
        """覆写给对象的属性赋值的方法，当属性值变化时，自动向观察者发送通知"""
        self._var_theta = float(new_value)  # 属性变化
        self.notify_observers()  # 自动向观察者发送通知, 参数已改变

    @property
    def var_vol(self):
        return self._var_vol

    @var_vol.setter
    def var_vol(self, new_value):
        """覆写给对象的属性赋值的方法，当属性值变化时，自动向观察者发送通知"""
        self._var_vol = float(new_value)  # 属性变化
        self.notify_observers()  # 自动向观察者发送通知, 参数已改变

    @property
    def var_rho(self):
        return self._var_rho

    @var_rho.setter
    def var_rho(self, new_value):
        """覆写给对象的属性赋值的方法，当属性值变化时，自动向观察者发送通知"""
        self._var_rho = float(new_value)  # 属性变化
        self.notify_observers()  # 自动向观察者发送通知, 参数已改变



#!/user/bin/env python
# -*- coding: utf-8 -*-
"""
Copyright (C) 2024 Galaxy Technologies
Licensed under the Apache License, Version 2.0
"""
import numpy as np
from scipy.stats import norm
from pricelib.common.utilities.enums import CallPut, PaymentType, ExerciseType
from pricelib.common.time import global_evaluation_date
from pricelib.common.pricing_engine_base import AnalyticEngine


class AnalyticAssetOrNothingEngine(AnalyticEngine):
    """二元(数字)期权-资产或无(asset or nothing)-闭式解定价引擎
    由于到期获得的payoff是标的资产多头，因此rebate固定为0"""

    # pylint: disable=invalid-name
    def calc_present_value(self, prod, t=None, spot=None):
        """计算现值
        Args:
            prod: Product产品对象
            t: datetime.date，估值日; 如果是None，则使用全局估值日globalEvaluationDate
            spot: float，估值日标的价格，如果是None，则使用随机过程的当前价格
        Returns: float，现值
        """
        assert prod.rebate == 0, "资产或无期权(asset or nothing)的rebate必须为0"
        calculate_date = global_evaluation_date() if t is None else t
        tau = prod.trade_calendar.business_days_between(calculate_date, prod.end_date) / prod.t_step_per_year
        if spot is None:
            spot = self.process.spot()

        r = self.process.interest(tau)
        q = self.process.div(tau)
        vol = self.process.vol(tau, spot)

        numerator = r - q + vol ** 2 / 2
        denominator = vol * np.sqrt(tau)
        mu = (r - q) / vol ** 2 - 0.5
        # 欧式
        if prod.exercise_type == ExerciseType.European:
            K_S = prod.strike / spot
            return spot * np.exp(-q * tau) * norm.cdf(
                prod.callput.value * (numerator * tau - np.log(K_S) / denominator))
        if prod.exercise_type == ExerciseType.American:
            if prod.discrete_obs_interval is not None:
                # 均匀离散观察障碍期权，M. Broadie, P. Glasserman, S.G. Kou(1997) 在连续障碍期权解析解上加调整项，调整障碍价格水平
                # 指数上的beta = -zeta(1/2) / sqrt(2*pi) = 0.5826, 其中zeta是黎曼zeta函数
                if prod.callput == CallPut.Call:  # 向上
                    strike = prod.strike * np.exp(0.5826 * vol * np.sqrt(prod.discrete_obs_interval))
                else:  # self.callput == CallPut.Put:  # 向下
                    strike = prod.strike * np.exp(-0.5826 * vol * np.sqrt(prod.discrete_obs_interval))
            else:  # 连续观察
                strike = prod.strike
            K_S = strike / spot
            # 美式，立即支付
            if prod.payment_type == PaymentType.Hit:
                lambda_ = np.sqrt(mu ** 2 + 2 * r / vol ** 2)
                return strike * (K_S ** (mu + lambda_) * norm.cdf(
                    -prod.callput.value * (np.log(K_S) / denominator + lambda_ * denominator)) + K_S ** (
                                         mu - lambda_) * norm.cdf(
                    -prod.callput.value * (np.log(K_S) / denominator - lambda_ * denominator)))
            # 美式，到期支付
            # else:  self.payment_type == PaymentType.Expire:
            return spot * np.exp(-q * tau) * (
                    norm.cdf((numerator * tau - np.log(K_S)) / denominator * prod.callput.value) +
                    K_S ** (2 * mu + 2) * norm.cdf(
                        -(numerator * tau + np.log(K_S)) / denominator * prod.callput.value))
        # 观察方式既不是欧式也不是美式
        raise NotImplementedError(
            f"ExerciseType {prod.exercise_type.value}不支持的观察方式，仅支持欧式观察或美式观察.")



#!/user/bin/env python
# -*- coding: utf-8 -*-
"""
Copyright (C) 2024 Galaxy Technologies
Licensed under the Apache License, Version 2.0
"""
import math
import numpy as np
from scipy.stats import norm, multivariate_normal
from pricelib.common.utilities.enums import CallPut, ExerciseType
from pricelib.common.processes import StochProcessBase
from pricelib.common.time import global_evaluation_date
from pricelib.common.pricing_engine_base import AnalyticEngine


class AnalyticVanillaAmEngine(AnalyticEngine):
    """美式香草期权-解析解定价引擎
    1. BAW - Barone-Adesi and Whaley(1987)方法
    2. Bjerksund&Stensland2002 - Bjerksund and Stensland(2002)方法"""

    # pylint: disable=invalid-name
    def __init__(self, stoch_process: StochProcessBase = None, an_method="BAW", *,
                 s=None, r=None, q=None, vol=None):
        """构造函数
        Args:
            stoch_process: 随机过程对象
            an_method: 美式期权解析解方法："BAW",或 "Bjerksund&Stensland2002"
        在未设置stoch_process时，(stoch_process=None)，会默认创建BSMprocess，需要输入以下变量进行初始化
            s: float，标的价格
            r: float，无风险利率
            q: float，分红/融券率
            vol: float，波动率
        """
        super().__init__(stoch_process=stoch_process, s=s, r=r, q=q, vol=vol)
        self.an_method = an_method
        # 以下属性指向需要定价的产品，由calc_present_value方法设置
        self.prod = None
        # 以下为计算过程的中间变量
        self.spot = None
        self.T = None
        self.vol = None
        self.r = 0
        self.q = 0
        self.q_star = None
        self.beta = 1
        self.b = 0
        self.strike = None  # Bjerksund and Stensland(2002)方法计算Put时需要颠倒strike和spot，此变量临时存储strike

    # pylint: disable=invalid-name
    def calc_present_value(self, prod, t=None, spot=None):
        """计算现值
        Args:
            prod: Product产品对象
            t: datetime.date，估值日; 如果是None，则使用全局估值日globalEvaluationDate
            spot: float，估值日标的价格，如果是None，则使用随机过程的当前价格
        Returns: float，现值
        """
        assert prod.exercise_type == ExerciseType.American, "Error: 美式期权的行权类型只能是American"
        self.prod = prod
        calculate_date = global_evaluation_date() if t is None else t
        self.T = (prod.end_date - calculate_date).days / prod.annual_days.value
        if spot is None:
            spot = self.process.spot()
        self.spot = spot
        self.r = self.process.interest(self.T)
        self.q = self.process.div(self.T)
        self.vol = self.process.vol(self.T, spot)
        if self.q == 0 and prod.callput == CallPut.Call:
            # 如果没有分红，看涨期权的美式期权价值等于欧式期权价值
            return self.BSM_value(self.spot)
        if self.an_method == "Bjerksund&Stensland2002":
            price = self._Bjerksund_Stensland_2002()
        elif self.an_method == "BAW":
            price = self._BAW_1987()
        else:
            raise ValueError(f"{self.an_method} - 不支持的美式期权解析解方法")

        price = max(price, self.BSM_value(self.spot))
        return price

    # pylint: disable=invalid-name
    def d1(self, S):
        """BSM d1   S: 标的价格"""
        d1_value = (np.log(S / self.prod.strike) + (self.r - self.q + 0.5 * self.vol ** 2) * self.T) / (
                self.vol * np.sqrt(self.T))
        return d1_value

    # pylint: disable=invalid-name
    def d2(self, S):
        """BSM d2   S: 标的价格"""
        d2_value = self.d1(S) - self.vol * np.sqrt(self.T)
        return d2_value

    # pylint: disable=invalid-name
    def BSM_value(self, S):
        """BSM公式欧式香草期权现值    S: 标的价格"""
        return self.prod.callput.value * (
                S * math.exp(-self.q * self.T) * norm.cdf(self.prod.callput.value * self.d1(S))
                - self.prod.strike * math.exp(-self.r * self.T) * norm.cdf(
            self.prod.callput.value * self.d2(S)))

    # pylint: disable=invalid-name, missing-docstring
    def _BAW_1987(self):
        """Barone-Adesi and Whaley (1987)近似解"""

        def A(S_star):
            return self.prod.callput.value * (S_star / self.q_star * (
                    1 - math.exp(-self.q * self.T) * norm.cdf(self.prod.callput.value * self.d1(S_star))))

        def cal_S_star(S):
            LHS = self.prod.callput.value * (S - self.prod.strike)
            RHS = self.BSM_value(S) + A(S)
            n_iter = 0
            while abs(LHS - RHS) / self.prod.strike > 1e-6 and n_iter < 100:
                nom_b_1 = self.prod.callput.value * math.exp(-self.q * self.T) * norm.cdf(
                    self.prod.callput.value * self.d1(S)) * (1 - 1 / self.q_star)
                nom_b_2 = (1 - self.prod.callput.value * math.exp(-self.q * self.T) * norm.cdf(
                    self.prod.callput.value * self.d1(S)) / (
                                   self.vol * math.sqrt(self.T))) / self.q_star
                b = nom_b_1 + self.prod.callput.value * nom_b_2
                S = (self.prod.strike + self.prod.callput.value * RHS - self.prod.callput.value * b * S) / (
                        1 - self.prod.callput.value * b)
                LHS = self.prod.callput.value * (S - self.prod.strike)
                RHS = self.BSM_value(S) + A(S)
                n_iter += 1
            return S

        M = 2 * self.r / self.vol ** 2
        N = 2 * (self.r - self.q) / self.vol ** 2
        K = 1 - math.exp(-self.r * self.T)
        self.q_star = (-(N - 1) + self.prod.callput.value * math.sqrt((N - 1) ** 2 + 4 * M / K)) * 0.5

        s_inf = self.prod.strike / (1 - 2 / (N + 1 + self.prod.callput.value * ((N - 1) ** 2 + 4 * M) ** 0.5))
        h = -((self.r - self.q) * self.T + self.prod.callput.value * 2 * self.vol * self.T ** 0.5) * (
                self.prod.strike / (s_inf - self.prod.strike))

        if self.prod.callput == CallPut.Call:
            S_star = self.prod.strike + (s_inf - self.prod.strike) * (1 - math.e ** h)
            S_star = cal_S_star(S_star)
            if self.spot >= S_star:
                return self.spot - self.prod.strike
            # else:
            return self.BSM_value(self.spot) + A(S_star) * (self.spot / S_star) ** self.q_star
        if self.prod.callput == CallPut.Put:
            S_star = s_inf - (s_inf - self.prod.strike) * (1 - math.e ** h)
            S_star = cal_S_star(S_star)
            if self.spot <= S_star:
                return self.prod.strike - self.spot
            # else:
            return self.BSM_value(self.spot) + A(S_star) * (self.spot / S_star) ** self.q_star
        raise ValueError("Error: CallPut类型错误！")

    def _Bjerksund_Stensland_2002(self):
        """Bjerksund Stensland (2002)的美式期权近似解
            看涨期权，直接计算；
            看跌期权，公式中需要调换标的价格和执行价的值"""
        self.b = self.r - self.q
        assert self.b != 0, "Error: Bjerksund_Stensland_2002不支持 r-q=0 的情形！"
        if self.prod.callput is CallPut.Call:
            self.strike = self.prod.strike
            price = self._Bjerksund_Stensland_2002_am_call_value(self.r, self.b)
        else:  # self.callput is CallPut.Put:
            self.strike, self.spot = self.spot, self.prod.strike
            price = self._Bjerksund_Stensland_2002_am_call_value(self.q, -self.b)
            self.strike, self.spot = self.prod.strike, self.strike
        return price

    # pylint: disable=invalid-name, too-many-locals, missing-docstring
    def _Bjerksund_Stensland_2002_am_call_value(self, r, b):
        """
        Bjerksund and Stensland(2002)近似解 - 美式看涨期权
        Args:
            r: 无风险利率
            b: r-q，无风险利率-分红率
        Returns: 美式期权现值
        """

        def B_inf():
            return self.beta / (self.beta - 1) * self.strike

        def B0():
            return np.maximum(self.strike, r / (r - b) * self.strike)

        def M(a, b, rho):  # 标准二元分布函数的累积分布函数
            mean = np.array([0, 0])
            cov = np.array([[1, rho], [rho, 1]])
            return multivariate_normal(mean, cov).cdf([a, b])

        def hold_time(t):
            return -(b * t + 2 * self.vol * math.sqrt(t)) * (self.strike ** 2 / ((B_inf() - B0()) * B0()))

        def barrier_price(t):
            return B0() + (B_inf() - B0()) * (1 - np.exp(hold_time(t)))

        def alpha(x):
            return (x - self.strike) * x ** -self.beta

        def phi(S, T, gamma, H, X):
            d1 = -(np.log(S / H) + (b + (gamma - 0.5) * (self.vol ** 2)) * T) / (self.vol * np.sqrt(T))
            d2 = -(np.log(X ** 2 / (S * H)) + (b + (gamma - 0.5) * (self.vol ** 2)) * T) / (self.vol * np.sqrt(T))
            lamb = -r + gamma * b + 0.5 * gamma * (gamma - 1) * (self.vol ** 2)
            kappa = (2 * b) / (self.vol ** 2) + (2 * gamma - 1)
            return np.exp(lamb * T) * (S ** gamma) * (norm.cdf(d1) - ((X / S) ** kappa) * norm.cdf(d2))

        def psi(S, big_T, gamma, H, big_X, small_X, small_T):
            lamb = -r + gamma * b + 0.5 * gamma * (gamma - 1) * (self.vol ** 2)
            kappa = 2 * b / (self.vol ** 2) + (2 * gamma - 1)

            small_d1 = -(np.log(S / small_X) + (b + (gamma - 0.5) * (self.vol ** 2)) * small_T) / (
                    self.vol * np.sqrt(small_T))
            small_d2 = -(np.log((big_X ** 2) / (S * small_X)) + (b + (gamma - 0.5) * (self.vol ** 2)) * small_T) / (
                    self.vol * np.sqrt(small_T))
            small_d3 = -(np.log(S / small_X) - (b + (gamma - 0.5) * (self.vol ** 2)) * small_T) / (
                    self.vol * np.sqrt(small_T))
            small_d4 = -(np.log((big_X ** 2) / (S * small_X)) - (b + (gamma - 0.5) * (self.vol ** 2)) * small_T) / (
                    self.vol * np.sqrt(small_T))
            big_d1 = -(np.log(S / H) + (b + (gamma - 0.5) * (self.vol ** 2)) * big_T) / (self.vol * np.sqrt(big_T))
            big_d2 = -(np.log((big_X ** 2) / (S * H)) + (b + (gamma - 0.5) * (self.vol ** 2)) * big_T) / (
                    self.vol * np.sqrt(big_T))
            big_d3 = -(np.log((small_X ** 2) / (S * H)) + (b + (gamma - 0.5) * (self.vol ** 2)) * big_T) / (
                    self.vol * np.sqrt(big_T))
            big_d4 = -(np.log((S * (small_X ** 2) / (H * (big_X ** 2)))) + (
                    b + (gamma - 0.5) * (self.vol ** 2)) * big_T) / (self.vol * np.sqrt(big_T))

            return np.exp(lamb * big_T) * (S ** gamma) * (M(small_d1, big_d1, np.sqrt(small_T / big_T))
                                                          - (((big_X / S) ** kappa) * M(small_d2, big_d2,
                                                                                        np.sqrt(small_T / big_T)))
                                                          - ((small_X / S) ** kappa) * M(small_d3, big_d3,
                                                                                         -np.sqrt(small_T / big_T))
                                                          + ((small_X / big_X) ** kappa) * M(small_d4, big_d4,
                                                                                             -np.sqrt(small_T / big_T)))

        self.beta = (0.5 - b / self.vol ** 2) + math.sqrt((b / self.vol ** 2 - 0.5) ** 2 + 2 * r / self.vol ** 2)

        t_1 = 0.5 * (math.sqrt(5) - 1) * self.T
        barrier_1 = barrier_price(self.T - t_1)
        barrier_2 = barrier_price(self.T)

        if self.spot > barrier_2:
            return self.spot - self.strike
        # else:
        div1 = (alpha(barrier_2) * self.spot ** self.beta
                - alpha(barrier_2) * phi(self.spot, t_1, self.beta, barrier_2, barrier_2))
        option1 = (phi(self.spot, t_1, 1.0, barrier_2, barrier_2)
                   - phi(self.spot, t_1, 1.0, barrier_1, barrier_2)
                   - self.strike * phi(self.spot, t_1, 0.0, barrier_2, barrier_2)
                   + self.strike * phi(self.spot, t_1, 0.0, barrier_1, barrier_2))
        div2 = (alpha(barrier_1) * phi(self.spot, t_1, self.beta, barrier_1, barrier_2)
                - alpha(barrier_1) * psi(self.spot, self.T, self.beta, barrier_1, barrier_2, barrier_1, t_1))
        option2 = (psi(self.spot, self.T, 1.0, barrier_1, barrier_2, barrier_1, t_1)
                   - psi(self.spot, self.T, 1.0, self.strike, barrier_2, barrier_1, t_1)
                   - self.strike * psi(self.spot, self.T, 0.0, barrier_1, barrier_2, barrier_1, t_1)
                   + self.strike * psi(self.spot, self.T, 0.0, self.strike, barrier_2, barrier_1, t_1))
        return div1 + option1 + div2 + option2
